"""Install

Routine to install a parser/writer/converter style.

"""

import os
import re
import sys
import site
import shutil
import urllib2
import zipfile
import textwrap
import distutils.dir_util
import distutils.errors
import os.path as pth
from glob import iglob, glob
from imp import load_source
from pkg_resources import parse_version
from lexor.util.logging import L
from lexor.command import config, disp, LexorError, exec_cmd
from lexor.command.cloud import cloud_request
from lexor.util import github


DESC = """
Install a parser/writer/converter style.

"""
GITHUB_PATTERN = re.compile(
    r'(?:@|://)github.com[:/]([^/\s]+?)/([^/\s]+?)(?:\.git)?/?$',
    re.IGNORECASE
)


def add_parser(subp, fclass):
    """Add a parser to the main subparser. """
    tmpp = subp.add_parser('install', help='install a style',
                           formatter_class=fclass,
                           description=textwrap.dedent(DESC))
    tmpp.add_argument('style', type=str, nargs="?",
                      help='name of style to install')
    tmpp.add_argument('-u', '--user', action='store_true',
                      help='install in user-site')
    tmpp.add_argument('-g', '--global', action='store_true',
                      help='install globably, requires sudo')
    tmpp.add_argument('--path', type=str, default=None,
                      help='specify the installation path')
    tmpp.add_argument('-s', '--save', action='store_true',
                      help='save dependency')


def decompose(endpoint):
    """Parse an endpoint. Extracted from

        https://github.com/bower/endpoint-parser/blob/master/index.js

    """
    exp = '^(?:([\w\-]|(?:[\w\.\-]+[\w\-])?)=)?([^\|#]+)(?:#(.*))?$'
    matches = re.match(exp, endpoint)
    if not matches:
        raise LexorError('invalid endpoint %r' % endpoint)
    matches = matches.groups()
    name = '' if not matches[0] else matches[0].strip()
    source = '' if not matches[1] else matches[1].strip()
    target = '' if not matches[2] else matches[2].strip()
    is_wild_card = not target or target == '*' or target == 'latest'
    return {
        'name': name,
        'source': source,
        'target': '*' if is_wild_card else target
    }


def get_org_repo_pair(url):
    """Parse a github url returning the organization/user and the
    repo name. """
    matches = re.search(GITHUB_PATTERN, url)
    if not matches:
        return None
    matches = matches.groups()
    org = '' if not matches[0] else matches[0].strip()
    repo = '' if not matches[1] else matches[1].strip()
    return {
        'org': org,
        'repo': repo
    }


def _get_key_typedir(info, install_dir):
    """Helper function for install_style. """
    if info['to_lang']:
        key = '%s.%s.%s.%s' % (info['lang'], info['type'],
                               info['to_lang'], info['style'])
        typedir = '%s/%s.%s.%s'
        typedir = typedir % (install_dir, info['lang'], info['type'],
                             info['to_lang'])
    else:
        key = '%s.%s.%s' % (info['lang'], info['type'], info['style'])
        typedir = '%s/%s.%s'
        typedir = typedir % (install_dir, info['lang'], info['type'])
    return key, typedir


def install_style(style, install_dir):
    """Install a given style to the install_dir path. """
    if not style.startswith('/'):
        raise NameError('`style` is not an absolute path')
    if not install_dir.startswith('/'):
        raise NameError('`install_dir` is not an absolute path')

    mod = load_source('tmp_mod', style)
    info = mod.INFO
    key, typedir = _get_key_typedir(info, install_dir)
    L.info('module %r will be installed in %r', key, typedir)

    if not pth.exists(typedir):
        try:
            os.makedirs(typedir)
        except OSError:
            msg = 'OSError: unable to create directory %r. ' % typedir
            msg += 'Did you `sudo`?\n'
            raise LexorError(msg)

    moddir = pth.splitext(style)[0]
    base, name = pth.split(moddir)
    if base == '':
        base = '.'

    src = '%s/%s.py' % (base, name)
    dest = '%s/%s.py' % (typedir, name)
    disp('writing %r ... ' % dest)
    try:
        L.info('copying main file %r', src)
        shutil.copyfile(src, dest)
    except OSError:
        L.error('OSError: unable to copy file. Did you `sudo`?')
    disp('done\n')

    src = '%s/%s' % (base, name)
    if pth.exists(src):
        dest = '%s/%s' % (typedir, name)
        disp('writing %s ... ' % dest)
        try:
            L.info('copying auxiliary directory %r', src)
            distutils.dir_util.copy_tree(src, dest)
        except distutils.errors.DistutilsFileError as err:
            L.warn('DistutilsFileError: %r', err.message)
        disp('done\n')

    L.info('compiling modules ...')
    src = '%s/%s.py' % (typedir, name)
    load_source('tmp_mod', src)
    L.info('    - %r', src)

    src = '%s/%s/*.py' % (typedir, name)
    for path in iglob(src):
        load_source('tmp_mod', path)
        L.info('    - %r', path)

    msg = '  -> %r v%s has been installed in %r\n'
    disp(msg % (key, info['ver'], install_dir))
    return key, info['ver']


def download_file(url, base='.'):
    """Download a file. """
    try:
        L.info('downloading %s', url)
        response = urllib2.urlopen(url)
        local_name = '%s/tmp_%s' % (base, pth.basename(url))
        with open(local_name, "wb") as local_file:
            local_file.write(response.read())
        return local_name
    except urllib2.HTTPError, err:
        msg = 'HTTP Error [%r]: %r' % (err.code, url)
        L.error(msg)
        raise LexorError(msg)
    except urllib2.URLError, err:
        msg = 'URL Error -> %r: %r' % (err.reason, url)
        L.error(msg)
        raise LexorError(msg)


def unzip_file(local_name):
    """Extract the contents of a zip file. """
    zfile = zipfile.ZipFile(local_name)
    dirname = zfile.namelist()[0].split('/')[0]
    zfile.extractall()
    return dirname


def _get_install_dir(arg):
    """Get the installation directory. """
    if arg['user']:
        try:
            install_dir = '%s/lib/lexor_modules' % site.getuserbase()
        except AttributeError:
            install_dir = 'lib/lexor_modules'
        L.info('user installation: %r', install_dir)
    elif arg['global']:
        install_dir = '%s/lib/lexor_modules' % sys.prefix
        L.info('global installation: %r', install_dir)
    elif arg['path']:
        install_dir = pth.abspath(arg['path'])
        L.info('custom installation: %r', install_dir)
    else:
        install_dir = pth.join(pth.abspath('.'), 'lexor_modules')
        L.info('default installation: %r', install_dir)
    return install_dir


def _is_local_installation(style_file):
    """Check if the argument exists locally. """
    if '.py' not in style_file:
        style_file = '%s.py' % style_file
    if pth.exists(style_file):
        return pth.abspath(style_file)
    return None


def _parse_key(key):
        info = key.split('.')
        if len(info) == 3:
            return {
                'lang': info[0],
                'type': info[1],
                'style': info[2]
            }
        elif len(info) == 4:
            return {
                'lang': info[0],
                'type': info[1],
                'to_lang': info[2],
                'style': info[3]
            }
        else:
            types = ['lang.type.style', 'lang.type.to_lang.style']
            msg = 'invalid module, try %r' % types
            raise LexorError(msg)


def install_url(style, url, install_dir, version=''):
    """Download and install a zip file"""
    if version and version[0] != 'v':
        version = 'v{0}'.format(version)
    msg = 'installing %s %s ... \n'
    disp(msg % (style, version))
    try:
        local_name = download_file(url)
    except LexorError:
        L.error('something went wrong while downloading ...')
        return
    dirname = unzip_file(local_name)
    source = glob('%s/*.py' % dirname)
    key, ver = install_style(pth.abspath(source[0]), install_dir)
    L.info('removing %r and %r ...', local_name, dirname)
    os.remove(local_name)
    shutil.rmtree(dirname)
    L.info('clean up complete')
    return key, ver


def github_resolver(install_dir, source, target):
    return git_remote_resolver(install_dir, source, target)


def git_remote_resolver(install_dir, source, target):
    L.info('creating temporary directory ...')
    os.mkdir('tmp')
    os.chdir('tmp')
    exec_cmd('git init')
    exec_cmd('git pull --tags %s' % source)
    if target != '*':
        exec_cmd('git checkout %s' % target)
    os.chdir('..')
    source = glob('tmp/*.py')
    info = install_style(pth.abspath(source[0]), install_dir)
    L.info('removing temporary directory ...')
    shutil.rmtree('tmp')
    L.info('clean up complete')
    return info


def url_resolver(install_dir, source, _):
    return install_url(source, source, install_dir)


def local_resolver(install_dir, source, _):
    L.info('installing local module: %r', source)
    return install_style(source, install_dir)


def shorthand_resolver(install_dir, source, target):
    org, repo = source.split('/')
    return _get_github_archive(install_dir, org, repo, target)


def registry_resolver(install_dir, source, target):
    match_parameters = _parse_key(source)
    L.info('searching registry for %r', source)
    response = cloud_request('match', match_parameters)
    if len(response) == 0:
        L.error('no matches found for %r', source)
        return
    if len(response) > 1:
        msg = 'there are %d matches, how did this happen?'
        L.error(msg % len(response))
        return
    info = response[0]
    return _get_github_archive(
        install_dir, info['user'], info['repo'], target
    )


def _get_github_archive(install_dir, org, repo, target):
    if target == '*':
        endpoint = '/repos/{org}/{repo}/tags'.format(
            org=org,
            repo=repo
        )
        response = github.get(endpoint)
        if not isinstance(response, list):
            if 'message' in response:
                msg = 'message from github: %s' % response['message']
                L.error(msg)
            else:
                L.error('unable to git proper response from github.')
            return
        sha = {}
        for item in response:
            key = item['name']
            if key[0] == 'v':
                key = key[1:]
            sha[key] = item['commit']['sha']
        versions = sha.keys()
        versions = sorted(versions, key=parse_version)
        target = sha[versions[-1]]
        L.info('found versions %r', versions)
    url = 'https://github.com/{org}/{repo}/archive/{target}.zip'
    url = url.format(org=org, repo=repo, target=target)
    return install_url(repo, url, install_dir)


def get_resolver(source):
    """Get the resolver for the source. """
    # Git Case: git git+ssh, git+http, git+https
    #           .git at the end (probably ssh shorthand)
    #            git@ at the start
    if re.match('^git(\+(ssh|https?))?://', source) \
            or re.search('\.git/?$', source) \
            or re.match('^git@', source):
        source = re.sub('^git\+', '', source)
        if get_org_repo_pair(source):
            return [github_resolver, source]
        return [git_remote_resolver, source]
    # url
    if re.match('^https?://', source):
        return [url_resolver, source]
    # local
    if pth.exists(source) or pth.exists('%s.py' % source):
        source = source if '.py' in source else '%s.py' % source
        return [local_resolver, pth.abspath(source)]
    # shorthand
    parts = source.split('/')
    if len(parts) == 2:
        return [shorthand_resolver, source]
    # registry
    return [registry_resolver, source]


def handle_argument(install_dir, param, save):
    dec_endpoint = decompose(param)
    source = dec_endpoint['source']
    target = dec_endpoint['target']
    resolver, source = get_resolver(source)
    info = resolver(install_dir, source, target)
    if save:
        if info is None:
            return
        cfg = config.read_config()
        endpt = dec_endpoint
        if endpt['target'] != '*':
            endpt['target'] = '#{0}'.format(endpt['target'])
        else:
            endpt['target'] = ''
        source = '%s%s' % (endpt['source'], endpt['target'])
        dep = 'dependencies'
        try:
            cfg[dep][info[0]] = source
        except KeyError:
            cfg.add_section(dep)
            cfg[dep][info[0]] = source
        config.write_config(cfg)


def run():
    """Run the command. """
    arg = vars(config.CONFIG['arg'])
    install_dir = _get_install_dir(arg)

    if arg['style']:
        return handle_argument(install_dir, arg['style'], arg['save'])
    else:
        cfg = config.get_cfg(['dependencies'])
        for key in cfg['dependencies']:
            handle_argument(
                install_dir,
                cfg['dependencies'][key],
                False
            )



"""
Provides the |Writer| object which defines the basic mechanism for
writing the objects defined in :mod:`lexor.core.elements`. This
involves using objects derived from the abstract class |NodeWriter|.

.. |Writer| replace:: :class:`~.Writer`
.. |NodeWriter| replace:: :class:`~.NodeWriter`
.. |Node| replace:: :class:`~lexor.core.node.Node`
.. |CharData| replace:: :class:`~lexor.core.elements.CharacterData`
.. |Element| replace:: :class:`~lexor.core.elements.Element`

"""
import re
from cStringIO import StringIO
from lexor.command.lang import get_style_module
from lexor.command import config
RE = re.compile(" ")


def _replacer(*key_val):
    """Helper function for replace.

    Source: <http://stackoverflow.com/a/15221068/788553>
    """
    replace_dict = dict(key_val)
    replacement_function = lambda match: replace_dict[match.group(0)]
    pattern = re.compile("|".join([re.escape(k) for k, _ in key_val]), re.M)
    return lambda string: pattern.sub(replacement_function, string)


def replace(string, *key_val):
    """Replacement of strings done in one pass. Example:

        >>> replace("a < b && b < c", ('<', '&lt;'), ('&', '&amp;'))
        'a &lt; b &amp;&amp; b &lt; c'

    Source: http://stackoverflow.com/a/15221068/788553

    """
    return _replacer(*key_val)(string)


def find_whitespace(line, start, lim):
    """Attempts to find the index of the first whitespace before
    `lim`, if its not found, then it looks ahead. """
    index = line.rfind(' ', start, lim+1)
    if index != -1:
        return index
    index = line.find(' ', lim)
    if index != -1:
        return index
    return len(line)


class NodeWriter(object):
    """A node writer is an object which writes a node in three steps:

    - `start`
    - `data/child`
    - `end`

    """

    def __init__(self, writer):
        """A |NodeWriter| needs to be initialized with a |Writer|
        object. If this method is to be overloaded then make sure
        that it only accepts one parameter: `writer`. This method is
        used by the |Writer| and it calls it with itself as the
        parameter."""
        self.writer = writer

    def write(self, string, split=False):
        """Writes the string to a file object. The file object is
        determined by the |Writer| object that initialized this
        object (`self`). """
        self.writer.write_str(string, split)

    def start(self, node):
        """Overload this method to write part of the |Node| object in
        the first encounter with the |Node|. """
        pass

    def data(self, node):
        """This method gets called only by |CharData| nodes and
        should be overloaded to write their attribute ``data``,
        otherwise it will write the node's data as it is. """
        self.writer.write_str(node.data)

    @classmethod
    def child(cls, _):
        """This method gets called for |Element| nodes that have
        children. If it is overwritten then it will not traverse
        through child nodes unless you return something other than
        ``None``.

        This method by default returns ``True`` so that the |Writer|
        can traverse through the child nodes. """
        return True

    def end(self, node):
        """Overload this method to write part of the |Node| object in
        the last encounter with the |Node|. """
        pass


class DefaultWriter(NodeWriter):
    """If the language does not define a |NodeWriter| for
    ``__default__`` in the ``MAPPING`` dictionary then the writer
    will use this |NodeWriter|."""

    def start(self, node):
        """Write the start of the node as a XML tag. """
        att = ' '.join(['%s="%s"' % (k, v) for k, v in node.items()])
        if att != '':
            self.write('<%s %s>' % (node.name, att))
        else:
            self.write('<%s>' % node.name)

    def end(self, node):
        """Write the end of the node as an XML end tag. """
        self.write('</%s>' % node.name)


# The default of 7 attributes for class is too restrictive.
# pylint: disable=R0902
class Writer(object):
    """To see the languages available to the writer see the
    :mod:`lexor.command.lang` module. """

    def __init__(self, lang='xml', style='default', defaults=None):
        """Create a new writer by specifying the language and the
        style in which `Node` objects will be written. """
        if defaults is None:
            defaults = dict()
        self.defaults = defaults
        self.style_module = None
        self._lang = lang
        self._style = style
        self._filename = None
        self._file = None  # Points to a file object
        self._nw = None    # Array of NodeWriters
        self._reload = True  # Create new NodeWriters

        self._raw = None
        self._wrap = None
        self._buffer = None
        self._break_hint = None
        self._indent = None
        self._indent_empty = None
        self.pos = None
        self.width = None

        self.root = None   # The node to be written
        self.prev_str = None  # Reference to the last string printed

    def __getitem__(self, name):
        """Return the specified |NodeWriter|. """
        return self._nw.get(name, self._nw['__default__'])

    @property
    def filename(self):
        """
        .. admonition:: Read-Only Property
            :class: note

            The name of the file to which a |Node| object was last
            written to."""
        return self._filename

    @property
    def language(self):
        """The language in which the |Writer| writes |Node| objects.
        """
        return self._lang

    @language.setter
    def language(self, value):
        """_lang setter method. """
        self._lang = value
        self._reload = True

    @property
    def indent(self):
        """The indentation at the beginning of each line. """
        return self._indent

    @indent.setter
    def indent(self, value):
        """_indent setter method. """
        self.flush_buffer(tail=False)
        self._indent = value

    @property
    def writing_style(self):
        """The style in which the |Writer| writes a |Node| object. """
        return self._style

    @writing_style.setter
    def writing_style(self, value):
        """_style setter method. """
        self._style = value
        self._reload = True

    @property
    def string_buffer(self):
        """The current string buffer. This is the string that will
        be printed after its length exceeds the writer's width. """
        return self._buffer

    @string_buffer.setter
    def string_buffer(self, value):
        """_indent setter method. """
        self._buffer = value

    def set(self, lang, style, defaults=None):
        """Set the language and style in one call. """
        self._style = style
        self._lang = lang
        self.defaults = defaults
        self._reload = True

    def __str__(self):
        """Attempts to retrieve the last written string. """
        if self._filename is None and self._file is not None:
            return self._file.getvalue()
        if self._filename is not None:
            tmp = open(self._filename, 'r')
            val = tmp.read()
            tmp.close()
            return val
        return None

    def _write_str(self, string):
        """Helper function for write_str. """
        if string != '':
            self.prev_str = string
            self._file.write(string)
            nlines = string.count('\n')
            self.pos[0] += nlines
            if nlines > 0:
                self.pos[1] = len(string) - string.rfind('\n')
            else:
                self.pos[1] += len(string)

    def write_str(self, string, split=False):
        """The write function is meant to be used with |Node|
        objects. Use this function to write simple strings while the
        file descriptor is open. """
        if self._raw:
            self._write_str(string)
            return
        if not self._wrap:
            if self._indent != '':
                lines = string.split('\n')
                if self.pos[1] == 1:
                    if lines[0] != '' or self._indent_empty:
                        lines[0] = self._indent + lines[0]
                for num in range(1, len(lines)):
                    if lines[num] != '' or self._indent_empty:
                        lines[num] = self._indent + lines[num]
                self._write_str('\n'.join(lines))
            else:
                self._write_str(string)
            return
        if split:
            self._break_hint.append(string)
        lines = string.split('\n')
        num = 0
        while num < len(lines) - 1:
            self._buffer += lines[num]
            self.normalize_buffer()
            self.flush_buffer()
            self._write_str('\n')
            num += 1
        self._buffer += lines[num]
        self.normalize_buffer()

    def flush_buffer(self, tail=True):
        """Empty the contents of the buffer. """
        if not tail and self._buffer.endswith(' '):
            self._buffer = self._buffer[:-1]
        if self.pos[1] == 1:
            if self._buffer.startswith(' '):
                self._buffer = self._buffer[1:]
            if self._buffer != '':
                self._write_str(self._indent + self._buffer)
            elif self._indent_empty:
                self._write_str(self._indent)
        else:
            self._write_str(self._buffer)
        self._buffer = ''

    def normalize_buffer(self):
        """The term normalize means that the length of the buffer
        will be less than or equal to the wrapping width. Anything
        that exceeds the limit will be flushed. """
        line = self._buffer
        indent = self._indent
        if self.pos[1] > 1:
            indent = ''
        limit = self.width - self.pos[1] - len(indent) + 1
        while len(line) > limit:
            start = 0
            if line[start] == ' ':
                start += 1
            end = find_whitespace(line, start, limit)
            while self._break_hint:
                index = line.find(self._break_hint[0], start)
                del self._break_hint[0]
                if index > -1 and index <= limit:
                    if end > limit or index > end:
                        end = index
            if end == len(line):
                break
            self._write_str(indent + line[start:end] + '\n')
            if line[end:end+1] == ' ':
                line = line[end+1:]
            else:
                line = line[end:]
            indent = self._indent
            limit = self.width - self.pos[1] - len(indent) + 1
        self._buffer = line

    def enable_wrap(self):
        """Use this to set the writing in wrapping mode. """
        self._wrap = True

    def disable_wrap(self):
        """Turn off wrapping. """
        self.flush_buffer()
        self._wrap = False

    def enable_raw(self):
        """Use this to set the writing in raw mode. """
        self.flush_buffer()
        self._raw = True

    def disable_raw(self):
        """Turn off raw mode. """
        self._raw = False

    def raw_enabled(self):
        """Determine if raw mode is enabled or not. """
        return self._raw

    def wrap_enabled(self):
        """Determine if wrap mode is enabled or not. """
        return self._wrap

    def endl(self, force=True, tot=1, tail=False):
        """Insert a new line character. By setting `force` to
        ``False`` you may omit inserting a new line character if the
        last character printed was already the new line character."""
        prev_str = self.last()
        self.flush_buffer(tail)
        if force or (not prev_str.endswith('\n') and
                     prev_str != self._indent):
            self._write_str('\n'*tot)

    def last(self):
        """Returns the last written string with the contents of the
        buffer. """
        if self.pos[1] == 1 and self._buffer.startswith(' '):
            return self.prev_str + self._buffer[1:]
        return self.prev_str + self._buffer

    def write(self, node, filename=None, mode='w'):
        """Write node to a file or string. To write to a string use
        the default parameters, otherwise provide a `filename`. If
        `filename` is provided you have the option of specifying the
        mode: ``'w'`` or ``'a'``.

        You may also provide a file you may have opened yourself in
        place of filename so that the writer writes to that file.

        Use the :meth:`__str__` function to retrieve the contents
        written to a string."""
        if isinstance(filename, file):
            # Check for stdout, stderr
            self._filename = file
            self._file = filename
        elif filename is None:
            # Check for StringIO
            self._filename = None
            if self._file is not None:
                self._file.close()
            self._file = StringIO()
        else:
            self._filename = filename
            self._file = open(filename, mode)
        self.root = node
        self._raw = True
        self._wrap = False
        self._buffer = ''
        self._break_hint = []
        self._indent = ''
        self._indent_empty = False
        self.pos = [1, 1]
        self.width = 70
        self.prev_str = '\n'
        if self._reload:
            self._set_node_writers(self._lang, self._style, self.defaults)
            self._reload = False
        self._set_node_writers_writer()
        if hasattr(self.style_module, 'pre_process'):
            self.style_module.pre_process(self, node)
        self._write(node)
        self.flush_buffer()
        if hasattr(self.style_module, 'post_process'):
            self.style_module.post_process(self, node)
        if isinstance(filename, file):
            pass
        elif filename is not None:
            self._file.close()

    def close(self):
        """Close the file. """
        if self._filename is not file:
            self._file.close()

    def _set_node_writers(self, lang, style, defaults=None):
        """Imports the correct module based on the language and
        style. """
        self.style_module = get_style_module('writer', lang, style)
        name = '%s-writer-%s' % (lang, style)
        config.set_style_cfg(self, name, defaults)
        self._nw = dict()
        self._nw['__default__'] = DefaultWriter(self)
        nw_obj = NodeWriter(self)
        self._nw['#document'] = nw_obj
        self._nw['#document-fragment'] = nw_obj
        self._nw['#text'] = nw_obj
        self._nw['#entity'] = nw_obj
        self._nw['#cdata-section'] = nw_obj
        str_key = list()
        for key, val in self.style_module.MAPPING.iteritems():
            if isinstance(val, str):
                str_key.append((key, val))
            else:
                self._nw[key] = val(self)
        for key, val in str_key:
            self._nw[key] = self._nw[val]

    def _set_node_writers_writer(self):
        """To be called before writing since the file will change. """
        for key in self._nw:
            self._nw[key].writer = self

    def _write_start(self, node):
        """To be called during tree traversal when node is first
        encountered. """
        self._nw.get(node.name, self._nw['__default__']).start(node)

    def _write_data(self, node):
        """To be called during tree traversal after _write_start if the
        node is a `CharacterData`. """
        self._nw.get(node.name, self._nw['__default__']).data(node)

    def _write_child(self, node):
        """To be called during tree traversal after _write_start if the
        node has children. """
        return self._nw.get(node.name, self._nw['__default__']).child(node)

    def _write_end(self, node):
        """To be called during tree traversal on last visit to node. """
        self._nw.get(node.name, self._nw['__default__']).end(node)

    def _get_direction(self, crt):
        """Returns the direction in which the traversal should go. """
        if hasattr(crt, 'data'):
            self._write_data(crt)
            self._write_end(crt)
            return 'r'
        elif crt.child:
            if self._write_child(crt) is None:
                return 'r'
            else:
                return 'd'
        else:
            self._write_end(crt)
            return 'r'

    def _write(self, root):
        """To be called during actual write function. """
        crt = root
        direction = None
        self._write_start(crt)
        if hasattr(crt, 'data'):
            self._write_data(crt)
            self._write_end(crt)
            return
        if crt.child:
            if self._write_child(crt) is None:
                return
            else:
                direction = 'd'
        else:
            self._write_end(crt)
            return
        while True:
            if direction is 'd':
                crt = crt.child[0]
            elif direction is 'r':
                if crt.next is None:
                    direction = 'u'
                    continue
                crt = crt.next
            elif direction is 'u':
                self._write_end(crt.parent)
                if crt.parent is root:
                    break
                if crt.parent.next is None:
                    crt = crt.parent
                    continue
                crt = crt.parent.next
            self._write_start(crt)
            direction = self._get_direction(crt)



"""
The core of lexor is divided among the modules in this package.

:|node|:
    Provides the most basic structure to create the document object
    model (DOM).

:|elements|:
    Here we define the basic structures to handle the information
    provided in files. Make sure to familiarize yourself with all the
    objects in this module to be able to write extensions for the
    |Parser|, |Converter| and |Writer|.

:|parser|:
    The parser module provides the |Parser| and the abstract class
    |NodeParser| which helps us write derived objects for future
    languages to parse.

:|converter|:
    The converter module provides the |Converter| and the abstract
    class |NodeConverter| which helps us copy a |Document| we
    want to convert to another language.

:|writer|:
    The writer module provides the |Writer| and the abstract class
    |NodeWriter| which once subclassed help us tell the |Writer|
    how to write a |Node| to a file object.


.. |node| replace:: :mod:`~lexor.core.node`
.. |Node| replace:: :mod:`~lexor.core.node.Node`
.. |elements| replace:: :mod:`~lexor.core.elements`
.. |Document| replace:: :mod:`~lexor.core.elements.Document`
.. |parser| replace:: :mod:`~lexor.core.parser`
.. |Parser| replace:: :mod:`~lexor.core.parser.Parser`
.. |NodeParser| replace:: :mod:`~lexor.core.parser.NodeParser`
.. |converter| replace:: :mod:`~lexor.core.converter`
.. |Converter| replace:: :mod:`~lexor.core.converter.Converter`
.. |NodeConverter| replace:: :mod:`~lexor.core.converter.NodeConverter`
.. |writer| replace:: :mod:`~lexor.core.writer`
.. |Writer| replace:: :mod:`~lexor.core.writer.Writer`
.. |NodeWriter| replace:: :mod:`~lexor.core.writer.NodeWriter`

"""

from lexor.core.node import Node
from lexor.core.elements import (
    CharacterData,
    Text,
    ProcessingInstruction,
    Comment,
    CData,
    Entity,
    DocumentType,
    Element,
    RawText,
    Void,
    Document,
    DocumentFragment,
)
from lexor.core.parser import (
    NodeParser,
    Parser,
)
from lexor.core.writer import (
    NodeWriter,
    Writer,
    replace,
)
from lexor.core.converter import (
    BaseLog,
    NodeConverter,
    Converter,
    get_converter_namespace,
)
from lexor.core.selector import Selector



"""
This module defines the elements of the document object model (DOM).
This implementation follows most of the recommendations of w3_.

.. _w3: http://www.w3.org/TR/2012/WD-dom-20121206/

Inheritance Tree
----------------

|   :class:`lexor.core.node.Node` (``__builtin__.object``)
|        :class:`.CharacterData`
|             :class:`.Text`
|             :class:`.ProcessingInstruction`
|             :class:`.Comment`
|             :class:`.CData`
|             :class:`.Entity`
|             :class:`.DocumentType`
|        :class:`.Element`
|             :class:`.RawText` (:class:`.Element`, :class:`.CharacterData`)
|             :class:`.Void`
|             :class:`.Document`
|                  :class:`.DocumentFragment`

----------------------------------------------------------------------

"""
import os
import sys
from lexor.core import Node
LC = sys.modules['lexor.core']
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-instance-attributes


class CharacterData(Node):
    """A simple interface to deal with strings. """

    __slots__ = ('data')

    def __init__(self, text=''):
        """Set the data property to the value of `text` and set its
        name to ``'#character-data'``. """
        Node.__init__(self)
        self.name = '#character-data'
        self.data = text

    @property
    def node_value(self):
        """Return or set the value of the node. This property is a
        wrapper for the ``data`` attribute. """
        return self.data

    @node_value.setter
    def node_value(self, value):
        """Setter function for data attribute. """
        self.data = value


class Text(CharacterData):
    """A node to represent a string object."""

    __slots__ = ()

    def __init__(self, text=''):
        """Call its base constructor and set its name to ``'#text'``.
        """
        CharacterData.__init__(self, text)
        self.name = '#text'

    def clone_node(self, _=True):
        """Return a new ``Text`` node with the same data content. """
        return Text(self.data)


class ProcessingInstruction(CharacterData):
    """Represents a "processing instruction", used to keep
    processor-specific information in the text of the document. """
    __slots__ = ('_target')

    def __init__(self, target, data=''):
        """Create a `Text` node with its `data` set to data. """
        CharacterData.__init__(self, data)
        self.name = target
        self._target = target

    @property
    def target(self):
        """The target of this processing instruction."""
        return self._target

    @target.setter
    def target(self, new_target):
        """Setter function. """
        self.name = new_target
        self._target = new_target

    def clone_node(self, _=True):
        """Returns a new PI with the same data content. """
        return ProcessingInstruction(self._target, self.data)


class Comment(CharacterData):
    """A node to store comments. """

    __slots__ = ('type')

    def __init__(self, data=''):
        """Create a comment node. """
        CharacterData.__init__(self, data)
        self.name = '#comment'
        self.type = None

    @property
    def comment_type(self):
        """Type of comment. This property is meant to help with
        documents that support different styles of comments. """
        return self.type

    @comment_type.setter
    def comment_type(self, comment_type):
        """Setter function for comment_type. """
        self.type = comment_type

    def clone_node(self, _=True):
        """Returns a new comment with the same data content. """
        node = Comment(self.data)
        node.type = self.type
        return node


class CData(CharacterData):
    """Although this node has been deprecated from the DOM_, it
    seems that xml still uses it.

    .. _DOM: https://developer.mozilla.org/en-US/docs/Web/API/Node.nodeType

    """

    __slots__ = ()

    def __init__(self, data=''):
        """Create a CDATA node and set the node name to
        ``'#cdata-section'``."""
        CharacterData.__init__(self, data)
        self.name = '#cdata-section'

    def clone_node(self, _=True):
        """Returns a new ``CData`` node with the same data content.
        """
        return CData(self.data)


class Entity(CharacterData):
    """From merriam-webster definition_:

    - *Something that exists by itself*.
    - *Something that is separate from other things*.

    This node acts in the same way as a :class:`.Text` node but it
    has one main difference. The data it contains should contain no
    white spaces. This node should be reserved for special characters
    or words that have different meanings across different languages.
    For instance in HTML you have the ``&amp;`` to represent ``&``.
    In LaTeX you have to type ``\\$`` to represent ``$``. Using this
    node will help you handle these Entities hopefully more
    efficiently than simply finding and replacing them in a Text node.

    .. _definition: http://www.merriam-webster.com/dictionary/entity

    """

    __slots__ = ()

    def __init__(self, text=''):
        """Create an ``Entity`` node and set the node name to
        ``#entity``."""
        CharacterData.__init__(self, text)
        self.name = '#entity'

    def clone_node(self, _=True):
        """Returns a new ``Entity`` with the same data content. """
        return Entity(self.data)


class DocumentType(CharacterData):
    """A node to store the doctype declaration. This node will not
    follow the specifications at this point (May 30, 2013). It will
    simply recieve the string in between ``<!doctype`` and ``>``.

    Specs: http://www.w3.org/TR/2012/WD-dom-20121206/#documenttype

    """
    __slots__ = ()

    def __init__(self, data=''):
        """Create a ``DocumentType`` node and set its name to
        ``#doctype``. """
        CharacterData.__init__(self, data)
        self.name = '#doctype'
        # The next properties should be obtained from data
        # The specs do not mention type, instead they mention name.
        # The node name here is #doctype so that it can be easily
        # identified. The doctype "name" as they refer is called type.

    def clone_node(self, _=True):
        """Returns a new doctype with the same data content. """
        node = DocumentType(self.data)
        return node


class Element(Node):
    """Node object configured to have child Nodes and attributes. """

    def __init__(self, name, data=None):
        """The parameter ``data`` should be a ``dict`` object. The
        element will use the keys and values to populate its
        attributes. You may modify the elements internal dictionary.
        However, this may unintentially overwrite the attributes
        defined by the ``__setitem__`` method. If you wish to add
        another attribute to the ``Element`` object use the
        convention of adding an underscore at the end of the
        attribute. i.e

        >>> strong = Element('strong')
        >>> strong.message_ = 'An internal message'
        >>> strong['message'] = 'Attribute message'

        """
        Node.__init__(self)
        if data is None:
            data = dict()
        self.__dict__.update(data)
        if isinstance(data, dict):
            self._order = data.keys()
        else:
            self._order = list()
            for key, _ in data:
                if key not in self._order:
                    self._order.append(key)
        self.name = name
        self.child = list()

    def __call__(self, selector):
        """Return a :class:`lexor.core.selector.Selector` object. """
        return LC.Selector(selector, self)

    def update_attributes(self, node):
        """Copies the attributes of the input node into the calling
        node. """
        for k in node:
            self.__dict__[k] = node.__dict__[k]
            if k not in self._order:
                self._order.append(k)

    def __getitem__(self, k):
        """Return the `k`-th child of this node if `k` is an integer.
        Otherwise return the attribute of name with value of `k`.

        >>> x.__getitem__(k) is x[k]
        True

        """
        if isinstance(k, str):
            return self.__dict__[k]
        if self.child:
            return self.child[k]
        return None

    def get(self, k, val=''):
        """Return the attribute of name with value of `k`."""
        return self.__dict__.get(k, val)

    def __setitem__(self, k, val):
        """Overloaded array operator. Appends or modifies an
        attribute. See its base method
        :meth:`lexor.core.node.Node.__setitem__` for documentation on
        when `val` is not string.

        >>> x.__setitem__(attname) = 'att' <==> x[attname] = 'att'

        """
        if isinstance(k, str):
            self.__dict__[k] = val
            if k not in self._order:
                self._order.append(k)
            if k == 'id' and self.owner:
                self.owner.id_dict[k] = self
        else:
            Node.__setitem__(self, k, val)

    def __delitem__(self, k):
        """Remove a child or attribute.

        >>> x.__delitem__(k) <==> del x[k]

        """
        if isinstance(k, str):
            self.__dict__.__delitem__(k)
            self._order.remove(k)
        else:
            Node.__delitem__(self, k)

    def __contains__(self, obj):
        """Return ``True`` if `obj` is a node and it is a child
        of this element or if `obj` is an attribute of this
        element. Return ``False`` otherwise.

        >>> x.__contains__(obj) == obj in x
        True

        """
        if isinstance(obj, Node):
            return self.child.__contains__(obj)
        else:
            return self._order.__contains__(obj)

    def contains(self, obj):
        """Unlike ``__contains__``, this method returns ``True`` if
        `obj` is any of the desendents of the node. """
        if obj.level < self.level + 1:
            return False
        while obj.level > self.level + 1:
            obj = obj.parent
            if obj is None:
                return False
        return obj in self

    def __iter__(self):
        """Iterate over the element attributes names.

        >>> for attribute_name in node: ...

        """
        for k in self._order:
            yield k

    @property
    def attlen(self):
        """The number of attributes. """
        return len(self._order)

    @property
    def attributes(self):
        """Return a list of the attribute names in the element. """
        return list(self._order)

    @property
    def values(self):
        """Return a list of the attribute values in the Element. """
        return [self.__dict__[k] for k in self._order]

    def attribute(self, index):
        """Return the name of the attribute at the specified index. """
        return self._order[index]

    def attr(self, index):
        """Return the value of the attribute at the specified index.
        """
        return self.__dict__[self._order[index]]

    def items(self):
        """return all the items. """
        return zip(self._order, self.values)

    def update(self, dict_):
        """update with the values of `dict_`. useful when the element
        is empty and you created an Attr object. then just update the
        values."""
        for key, val in dict_.items():
            self.__setitem__(key, val)

    def rename(self, old_name, new_name):
        """Renames an attribute. 

        >>> from lexor.core.elements import Element
        >>> node = Element('div')
        >>> node['att1'] = 'val1'
        >>> node
        div[0x10a090750 att1="val1"]:
        >>> node.rename('att1', 'new-att-name')
        >>> node
        div[0x10a090750 new-att-name="val1"]:
        
        """
        if isinstance(old_name, str):
            index = self._order.index(old_name)
        else:
            index = old_name  # Assume old_name
        self.__dict__[new_name] = self.__dict__[self._order[index]]
        del self.__dict__[self._order[index]]
        self._order[index] = new_name

    def clone_node(self, deep=False, normalize=True):
        """Returns a new node. When deep is True, it will clone also
        clone all the child nodes."""
        # May want to provide a node to which the clone will be
        # appended to. If this is done then we will not have to
        # traverse through all the elements of the node to adjust
        # the level of the child nodes when we move the node around
        node = Element(self.name)
        node.update_attributes(self)
        if deep is False or not self.child:
            return node
        crt = self
        crtcopy = node
        direction = 'd'
        while True:
            if direction is 'd':
                crt = crt.child[0]
                clone = crt.clone_node()
                crtcopy.append_child(clone)
            elif direction is 'r':
                if crt.next is None:
                    direction = 'u'
                    continue
                crt = crt.next
                clone = crt.clone_node()
                crtcopy.parent.append_child(clone)
            elif direction is 'u':
                crtcopy = crtcopy.parent
                if normalize:
                    crtcopy.normalize()
                if crt.parent is self:
                    break
                if crt.parent.next is None:
                    crt = crt.parent
                    continue
                crt = crt.parent.next
                clone = crt.clone_node()
                crtcopy.parent.append_child(clone)
            crtcopy = clone
            if crt.child:
                direction = 'd'
            else:
                direction = 'r'
        return node

    def get_elements_by_class_name(self, classname):
        """Return a list of all child elements which have all of the
        given class names. """
        nodes = []
        if not self.child:
            return nodes
        patterns = set([i.strip() for i in classname.split()])
        crt = self
        direction = 'd'
        while True:
            if direction is 'd':
                crt = crt.child[0]
            elif direction is 'r':
                if crt.next is None:
                    direction = 'u'
                    continue
                crt = crt.next
            elif direction is 'u':
                if crt.parent is self:
                    break
                if crt.parent.next is None:
                    crt = crt.parent
                    continue
                crt = crt.parent.next
            if isinstance(crt, Element) and 'class' in crt:
                crtclass = [i.strip() for i in crt['class'].split()]
                if patterns.issubset(set(crtclass)):
                    nodes.append(crt)
            if crt.child:
                direction = 'd'
            else:
                direction = 'r'
        return nodes

    def children(self, children=None, **keywords):
        """Set the elements children by providing a list of nodes or
        a string. If using a string then you may provide any of the
        following keywords to dictate how to parse and convert:

        - parser_style: ``'_'``
        - parser_lang: ``'html``
        - parser_defaults: ``None``,
        - convert_style: ``'_'``,
        - convert_from: ``None``,
        - convert_to: ``'html'``,
        - convert_defaults: ``None``,
        - convert: ``'false'``

        If no children are provided then it returns a string of the children
        written in plain html. To change this behavior provide the
        following keywords:

        - writer_style: ``'plain'``
        - writer_lang: ``'html``

        .. important::

            This requires the installation of lexor styles.

        """
        if children is None:
            lang = keywords.get('writer_lang', 'html')
            style = keywords.get('writer_style', 'plain')
            writer = LC.Writer(lang, style)
            if self.owner is not None and self.owner.defaults is not None:
                for var, val in self.owner.defaults.iteritems():
                    writer.defaults[var] = os.path.expandvars(str(val))
            for var, val in keywords.iteritems():
                writer.defaults[var] = os.path.expandvars(str(val))
            result = ''
            for child in self.child:
                writer.write(child)
                result += str(writer)
            writer.close()
            return result
        if isinstance(children, str):
            info = {
                'parser_style': '_',
                'parser_lang': 'html',
                'parser_defaults': None,
                'convert_style': '_',
                'convert_from': None,
                'convert_to': 'html',
                'convert_defaults': None,
                'convert': 'false'
            }
            for att in keywords:
                info[att] = keywords[att]
            parser = LC.Parser(info['parser_lang'],
                               info['parser_style'],
                               info['parser_defaults'])
            parser.parse(children)
            if info['convert'] == 'true' and info['convert_to'] is not None:
                if info['convert_from'] is None:
                    info['convert_from'] = info['parser_lang']
                converter = LC.Converter(info['convert_from'],
                                         info['convert_to'],
                                         info['convert_style'],
                                         info['convert_defaults'])
                converter.convert(parser.doc)
                children = converter.doc.pop()
                converter.log.pop()
            else:
                children = parser.doc
            children.temporary = True
        self.remove_children()
        self.extend_children(children)


class RawText(Element, CharacterData):
    """A few elements do not have children, instead they have data.
    Such elements exist in HTML: ``script``, ``title`` among
    others."""

    def __init__(self, name, data='', att=None):
        """You may provide `att` as a ``dict`` object. """
        CharacterData.__init__(self, data)
        Element.__init__(self, name, att)
        self.child = None

    def clone_node(self, deep=True, normalize=True):
        """Returns a new ``RawText`` element"""
        node = RawText(self.name)
        node.update_attributes(self)
        if deep is True:
            node.data = self.data
        return node


class Void(Element):
    """An element with no children. """

    def __init__(self, name, att=None):
        """You may provide `att` as a `dict` object. """
        Element.__init__(self, name, att)
        self.child = None

    def clone_node(self, _=True, normalize=True):
        """Returns a new ``Void`` element. """
        node = Void(self.name)
        node.update_attributes(self)
        return node


class Document(Element):
    """Contains information about the document that it holds. """

    def __init__(self, lang='xml', style='default'):
        """Creates a new document object and sets its name to
        ``#document``."""
        Element.__init__(self, '#document')
        self.level = -1
        self.owner = self
        self.lang = lang
        self.style = style
        self.uri_ = None
        self.defaults = None
        self.id_dict = dict()
        self.meta = dict()
        self.temporary = True

    def clone_node(self, deep=False, normalize=True):
        """Returns a new Document. Note: it does not copy
        the default values. """
        node = Document(self.lang, self.style)
        node.update_attributes(self)
        node.uri_ = self.uri_
        node.meta.update(self.meta)
        if deep is False or not self.child:
            return node
        clone = Element.clone_node(self, deep, normalize)
        clone.name = ''  # not a document
        node.extend_children(clone)
        return node

    @property
    def language(self):
        """The current document's language. This property is used by
        the writer to determine how to write the document.

        This property is a wrapper for the ``lang`` attribute. """
        return self.lang

    @language.setter
    def language(self, val):
        """Setter function for language. """
        self.lang = val

    @property
    def writing_style(self):
        """The current document's style. This property is used by
        the writer to determine how to write the document.

        This property is a wrapper for the ``style`` attribute.
        """
        return self.style

    @writing_style.setter
    def writing_style(self, val):
        """Docstring for setter. """
        self.style = val

    @property
    def uri(self):
        """The Uniform Resource Identifier. This property may become
        useful if the document represents a file. This property
        should be set by the a :class:`~lexor.core.parser.Parser`
        object telling us the location of the file that it parsed
        into the Document object. """
        return self.uri_

    @staticmethod
    def create_element(tagname, data=None):
        """Utility function to avoid having to import
        ``lexor.core.elements`` module. Returns an element object. """
        return Element(tagname, data)

    def get_element_by_id(self, element_id):
        """Return the first element, in tree order, within the
        document whose ID is element_id, or None if there is none. """
        return self.id_dict.get(element_id, None)


class DocumentFragment(Document):
    """Takes in an element and "steals" its children. This element
    should only be used as a temporary container. Note that the
    ``__str__`` method may not yield the expected results since all
    the function will do is use the ``__str__`` method in each of its
    children. First assign this object to an actual Document. """

    def __init__(self, lang='xml', style='default'):
        Document.__init__(self, lang, style)
        self.name = '#document-fragment'

    def append_child(self, new_child):
        """Adds the node new_child to the end of the list of children
        of this node. The children contained in a
        ``DocumentFragment`` only have a parent (the
        ``DocumentFragment``). As opposed as
        :meth:`lexor.core.node.Node.append_child` which also takes
        care of the ``prev`` and ``next`` attributes. """
        if isinstance(new_child, str):
            new_child = Text(new_child)
        elif not isinstance(new_child, Node):
            raise TypeError("Only Nodes can be appended.")
        if new_child.parent is not None:
            del new_child.parent[new_child.index]
        self.child.append(new_child)
        new_child.parent = self
        new_child.owner = self
        return new_child

    def __repr__(self):
        """
        >>> x.__repr__() <==> repr(x)

        """
        return ''.join([repr(node) for node in self.child])

    def __str__(self):
        """
        >>> x.__str__() <==> str(x)

        """
        return ''.join([str(node) for node in self.child])



"""
Routine to create an xml file with the documentation of a lexor
style.

"""
# pylint: disable=star-args
import re
import os
import sys
import textwrap
import inspect
import lexor
import os.path as pth
from imp import load_source
from lexor.core import elements as core
from lexor.command import config
from lexor.command.to import style_parameters


def xml_style(lang_str):
    """Parses a style string. """
    input_style = style_parameters(lang_str)
    if input_style['name'] == '':
        input_style['name'] = '_'
    return input_style


DEFAULTS = {
    'path': '.',
}
DESC = """
Generate an xml file with the documentation of a lexor language style.

"""


def add_parser(subp, fclass):
    """Add a parser to the main subparser. """
    tmpp = subp.add_parser('document', help='document a style',
                           formatter_class=fclass,
                           description=textwrap.dedent(DESC))
    tmpp.add_argument('style', type=xml_style, nargs='?',
                      help='the xml style to write the documentation')
    tmpp.add_argument('--path', type=str,
                      help='search for styles in this directory')
    tmpp.add_argument('--output-dir', type=str, default='', metavar="DIR",
                      help='writes to file in the specified dir if set')


def run():
    """Run the command. """
    arg = config.CONFIG['arg']
    dirpath, fname = check_filename(arg)

    moddir = os.path.splitext(fname)[0]
    base, _ = os.path.split(moddir)
    if base == '':
        base = '.'

    mod = load_source('tmp-module', fname)
    doc = core.Document()
    if not hasattr(mod, 'INFO') or 'lang' not in mod.INFO:
        filename = fname[:-3] + '.xml'
        doc.append_child(make_module_node(mod, "main"))
    else:
        info = mod.INFO
        if info['to_lang']:
            filename = '%s/lexor.%s.%s.%s.%s-%s.xml'
            filename = filename % (dirpath, info['lang'],
                                   info['type'], info['to_lang'],
                                   info['style'], info['ver'])
        else:
            filename = '%s/lexor.%s.%s.%s-%s.xml'
            filename = filename % (dirpath, info['lang'],
                                   info['type'], info['style'],
                                   info['ver'])
        modules = append_main(doc, mod)
        #for mod_name in modules:
        #    doc.append_child(make_module_node(modules[mod_name]))

    warn('Writing %s ... ' % filename)
    if arg.output_dir == '':
        filename = None
    try:
        if arg.style:
            doc.style = arg.style['name']
            doc.defaults = arg.style['params']
        lexor.write(doc, filename)
    except IOError:
        error("\nERROR: unable to write file.\n"
              "xml writer default style missing?\n")
    else:
        warn('done\n')


def export_object(obj):
    """Process objects before they are exported into an xml document.
    Objects such as modules are exported as links in the restructured
    text format."""
    if isinstance(obj, str):
        return obj
    if inspect.ismodule(obj):
        return ':ref:`%s`' % obj.__name__
    if isinstance(obj, list):
        return [export_object(x) for x in obj]
    if isinstance(obj, dict):
        return {k: export_object(obj[k]) for k in obj}
    return obj


def get_info_node(info):
    """Generate the info node. """
    node_info = core.Element('info')
    for key in info:
        if info[key] is None or key == 'path':
            continue
        node = core.Element('info-entry')
        node['key'] = key
        if key == 'description':
            node.append_child(core.CData(str(info[key])))
        else:
            node.append_child(core.Text(str(info[key])))
        node_info.append_child(node)
    return node_info


def get_defaults_node(obj):
    """Obtain defaults node. """
    defaults_node = core.Element('defaults')
    for ele in obj:
        node = core.Element('entry')
        node['key'] = ele
        node.append_child(core.CData(repr(obj[ele])))
        defaults_node.append_child(node)
    return defaults_node


def get_mapping_node(mapping, repository=None):
    """Generate the mapping node. """
    if repository is None:
        repository = list()
    modules = dict()
    mapping_node = core.Element('mapping')
    keys = sorted(mapping.keys())
    for ele in keys:
        if isinstance(mapping[ele], tuple):
            entry = core.Element('mapping-entry')
            entry['key'] = ele
            entry.append_child(
                core.Element('checker').append_child(
                    core.CData(repr(mapping[ele][0]))
                )
            )
            for mod in mapping[ele][1]:
                node = core.Element('processor')
                if isinstance(mod, str):
                    for val in repository:
                        if val.__name__ == mod:
                            mod = val
                            break
                mod_name = mod.__module__
                node['module'] = mod_name
                node['name'] = mod.__name__
                mapping_node.append_child(node)
                if mod_name not in modules:
                    modules[mod_name] = sys.modules[mod_name]
                entry.append_child(node)
            mapping_node.append_child(entry)
        else:
            node = core.Element('entry')
            node['key'] = ele
            mod = mapping[ele]
            if isinstance(mod, str):
                for val in repository:
                    if val.__name__ == mod:
                        mod = val
                        break
            if isinstance(mod, str):
                node['from-entry'] = mod
                mapping_node.append_child(node)
                continue
            mod_name = mod.__module__
            node['module'] = mod_name
            node['name'] = mod.__name__
            if mod_name not in modules:
                modules[mod_name] = sys.modules[mod_name]
            mapping_node.append_child(node)
    return modules, mapping_node


def separate_objects(mod, remove=None):
    """Given a module, it separates the objects into a dictionary. """
    info = {
        'class': [],
        'function': [],
        'module': [],
        'data': [],
        'other': [],
    }
    mod_dict = dict(mod.__dict__)
    if remove is None:
        remove = list()
    for ele in mod_dict:
        if ele in remove or ele[0] == '_':
            continue
        if inspect.isfunction(mod_dict[ele]):
            if mod_dict[ele].__module__ == mod.__name__:
                info['function'].append(mod_dict[ele])
            else:
                info['other'].append([ele, mod_dict[ele]])
        elif inspect.isclass(mod_dict[ele]):
            if mod_dict[ele].__module__ == mod.__name__:
                info['class'].append(mod_dict[ele])
            else:
                info['other'].append([ele, mod_dict[ele]])
        elif inspect.ismodule(mod_dict[ele]):
            info['module'].append([ele, mod_dict[ele]])
        elif inspect.isbuiltin(mod_dict[ele]):
            info['other'].append([ele, mod_dict[ele]])
        else:
            info['data'].append([ele, mod_dict[ele]])
    return info


def get_function_node(func):
    """Return a function node. """
    node = core.Element('function')
    node['name'] = func.__name__
    argspec = inspect.getargspec(func)
    node_argspec = core.Element('argspec')
    node_argspec['varargs'] = str(argspec[1])
    node_argspec['keywords'] = str(argspec[2])
    node_args = core.Element('args')
    for item in argspec[0]:
        node_args.append_child(core.Void('arg', {'name': item}))
    if argspec[3] is not None:
        largs = len(argspec[0])
        ldefs = len(argspec[3])
        num = 0
        for index in xrange(largs-ldefs, largs):
            node_args[index]['default'] = argspec[3][num]
            num += 1
    node_argspec.append_child(node_args)
    node.append_child(node_argspec)
    doc = inspect.getdoc(func)
    if doc is not None:
        node.append_child(
            core.Element('doc').append_child(
                core.CData(doc)
            )
        )
    # return core.Text(func.__name__)
    return node


def get_property_node(prop):
    """Return a property node. """
    node = core.Element('property')
    node['name'] = prop[0]
    doc = inspect.getdoc(prop[1])
    if doc is not None:
        node.append_child(
            core.Element('doc').append_child(
                core.CData(doc)
            )
        )
    return node


def get_member_node(member):
    """Return a property node. """
    node = core.Element('member')
    node['name'] = member.__name__
    return node


def full_class_name(cls):
    """Obtain the full class name. """
    if 'tmp' in cls.__module__:
        return cls.__name__
    return cls.__module__ + "." + cls.__name__


def _update_class_tree(tree, cls, mro):
    """Helper function for get_class_node. It updates the tree. """
    for name, val in inspect.getmembers(cls):
        for mod in mro:
            if name in mod.__dict__:
                if inspect.ismethod(val):
                    if val.im_self is None:
                        tree[mod]['method'].append(val)
                    else:
                        tree[mod]['cls_method'].append(val)
                elif inspect.isdatadescriptor(val):
                    if val.__class__.__name__ == 'property':
                        tree[mod]['property'].append([name, val])
                    else:
                        tree[mod]['member'].append(val)


def _update_node(node, tree, mro):
    """Helper function for get_class_node. Updates the node. """
    for index in xrange(len(mro)):
        tmp = core.Element('cls_method_block')
        tmp['from'] = full_class_name(mro[index])
        for func in tree[mro[index]]['cls_method']:
            tmp.append_child(get_function_node(func))
        if len(tmp) > 0:
            node.append_child(tmp)

        tmp = core.Element('method_block')
        tmp['from'] = full_class_name(mro[index])
        for func in tree[mro[index]]['method']:
            tmp.append_child(get_function_node(func))
        if len(tmp) > 0:
            node.append_child(tmp)

        tmp = core.Element('property_block')
        tmp['from'] = full_class_name(mro[index])
        for prop in tree[mro[index]]['property']:
            tmp.append_child(get_property_node(prop))
        if len(tmp) > 0:
            node.append_child(tmp)

        tmp = core.Element('member_block')
        tmp['from'] = full_class_name(mro[index])
        for member in tree[mro[index]]['member']:
            tmp.append_child(get_member_node(member))
        if len(tmp) > 0:
            node.append_child(tmp)


def get_class_node(cls):
    """Return a class node. Assumes that cls is a valid class. """
    node = core.Element('class')
    node['name'] = cls.__name__

    tmp = core.Element('bases')
    for base in cls.__bases__:
        tmp.append_child(
            core.Void('base', {'name': full_class_name(base)})
        )
    node.append_child(tmp)

    mro = inspect.getmro(cls)
    tmp = core.Element('mro')
    for base in mro:
        tmp.append_child(
            core.Void('class', {'name': full_class_name(base)})
        )
    node.append_child(tmp)

    doc = inspect.getdoc(cls)
    if doc is not None:
        node.append_child(
            core.Element('doc').append_child(
                core.CData(doc)
            )
        )

    tree = dict()
    for mod in mro:
        tree[mod] = {
            'cls_method': [],
            'method': [],
            'property': [],
            'member': [],
        }
    _update_class_tree(tree, cls, mro)
    _update_node(node, tree, mro)
    return node


def append_main(doc, mod):
    """Append the main module. Return a dictionary containing all the
    modules used in the style. """
    module = core.Element('module')
    module['name'] = 'main'

    module.append_child(
        core.Element('doc').append_child(
            core.CData(str(mod.__doc__))
        )
    )
    module.append_child(get_info_node(mod.INFO))
    
    if hasattr(mod, 'DEFAULTS'):
        module.append_child(get_defaults_node(mod.DEFAULTS))
    if hasattr(mod, 'REPOSITORY'):
        repository = mod.REPOSITORY
    else:
        repository = None
    modules, m_node = get_mapping_node(mod.MAPPING, repository)
    module.append_child(m_node)
    info = separate_objects(mod, ['INFO', 'DEFAULTS',
                                  'MAPPING', 'DESCRIPTION'])

    node = core.Element('classes')
    for cls in info['class']:
        node.append_child(get_class_node(cls))
    module.append_child(node)

    node = core.Element('functions')
    for func in info['function']:
        node.append_child(get_function_node(func))
    module.append_child(node)

    node = core.Element('data_block')
    for ele in info['data']:
        if ele[0][0] == '_':
            continue
        tmp = core.Element('data', {'name': ele[0]})
        if isinstance(ele[1], type(re.compile(''))):
            tmp.append_child(core.CData(repr(ele[1].pattern)))
        else:
            tmp.append_child(core.CData(repr(export_object(ele[1]))))
        node.append_child(tmp)
    if len(node) > 0:
        module.append_child(node)

    node = core.Element('imports')
    for ele in info['module']:
        node.append_child(
            core.Void('module', {'name': ele[0],
                                 'fullname': ele[1].__name__})
        )
    for ele in info['other']:
        node.append_child(
            core.Void('other', {'name': ele[0],
                                'fullname': full_class_name(ele[1]),
                                'type': ele[1].__class__.__name__})
        )
    module.append_child(node)
    doc.append_child(module)
    return modules


def make_module_node(mod, name=None):
    """Create a module node documenation. """
    module = core.Element('module')
    if name is None:
        name = mod.__name__
    module['name'] = name

    module.append_child(
        core.Element('doc').append_child(
            core.CData(str(mod.__doc__))
        )
    )

    info = separate_objects(mod)

    node = core.Element('classes')
    for cls in info['class']:
        node.append_child(get_class_node(cls))
    module.append_child(node)

    node = core.Element('functions')
    for func in info['function']:
        node.append_child(get_function_node(func))
    module.append_child(node)

    node = core.Element('data_block')
    for ele in info['data']:
        if ele[0][0] == '_':
            continue
        tmp = core.Element('data', {'name': ele[0]})
        if isinstance(ele[1], type(re.compile(''))):
            tmp.append_child(core.CData(repr(ele[1].pattern)))
        else:
            tmp.append_child(core.CData(repr(ele[1])))
        node.append_child(tmp)
    if len(node) > 0:
        module.append_child(node)

    node = core.Element('imports')
    for ele in info['module']:
        node.append_child(
            core.Void('module', {'name': ele[0],
                                 'fullname': ele[1].__name__})
        )
    for ele in info['other']:
        node.append_child(
            core.Void('other', {'name': ele[0],
                                'fullname': full_class_name(ele[1]),
                                'type': ele[1].__class__.__name__})
        )
    module.append_child(node)
    return module


def check_filename(arg):
    """Check if the inputfile exists. """
    cfg = config.get_cfg('document', DEFAULTS)
    root = cfg['lexor']['root']
    path = cfg['document']['path']

    fname = arg.inputfile
    if path[0] in ['/', '.']:
        dirpath = path
    else:
        dirpath = '%s/%s' % (root, path)

    if '.py' not in fname:
        fname = '%s.py' % fname
    if not pth.exists(pth.join(dirpath, fname)):
        error("ERROR: %r not found.\n" % (pth.join(dirpath, fname)))
    return dirpath, fname



"""Paste

Routine to append paste templates.

"""
# pylint: disable=W0142

import os
import textwrap
from lexor.command import config, LexorError

DESC = """
Paste a template to create a new style.

"""


def add_parser(subp, fclass):
    """Add a parser to the main subparser. """
    types = ['parser', 'writer', 'converter',
             'node-parser', 'node-writer', 'node-converter']
    tmpp = subp.add_parser('paste', help='paste a template',
                           formatter_class=fclass,
                           description=textwrap.dedent(DESC))
    tmpp.add_argument('lang', type=str,
                      help='language name')
    tmpp.add_argument('type', type=str, metavar='type', choices=types,
                      help='file type: ' + ', '.join(types))
    tmpp.add_argument('style', type=str,
                      help='style name')
    tmpp.add_argument('optional', nargs='*', default=None,
                      help="[to language] [auxilary filename]")


def make_style(base, type_, fmt):
    """Creates a new style file. """
    template = '%s/../core/templates/%s-style.txt' % (base, type_)
    content = open(template).read()
    sfile = '%s.py' % fmt['style']
    if os.path.exists(sfile):
        print 'Opening: %s' % sfile
    else:
        print 'Creating style: %s' % sfile
        with open(sfile, 'w') as wfile:
            wfile.write(content.format(**fmt))
    return sfile


def make_auxilary(base, type_, fmt, aux_type=''):
    """Creates a new node parser module. """
    tmp = '' if aux_type == '' else '-test'
    template = '%s/../core/templates/%s%s.txt' % (base, type_, tmp)
    content = open(template).read()
    tmp = '' if aux_type == '' else 'test_'
    npfile = '%s%s.py' % (tmp, fmt['np'])
    if os.path.exists(npfile):
        print 'Opening: %s' % npfile
    else:
        print 'Creating node processor: %s' % npfile
        with open(npfile, 'w') as wfile:
            wfile.write(content.format(**fmt))
    return npfile


def _get_option(array, index, msg):
    """Exit if array index is not accessible."""
    try:
        return array[index]
    except IndexError:
        raise LexorError(msg)


def run():
    """Run the command. """
    arg = config.CONFIG['arg']

    base = os.path.dirname(__file__)
    lang = arg.lang
    style = arg.style
    type_ = arg.type

    if 'converter' in type_:
        msg = 'converter needs to_lang.'
        tolang = _get_option(arg.optional, 0, msg)
    else:
        tolang = ''

    fmt = {
        'LANG': lang.upper(),
        'STYLE': style.upper(),
        'TOLANG': tolang.upper(),
        'lang': lang,
        'style': style,
        'tolang': tolang
    }

    if 'node' in type_:
        msg = 'provide name of auxilary file.'
        if 'converter' in type_:
            name = _get_option(arg.optional, 1, msg)
        else:
            name = _get_option(arg.optional, 0, msg)
        fmt['NP'] = name.upper()
        fmt['np'] = name
        npfile = make_auxilary(base, type_, fmt)
        # os.system('%s "%s" > /dev/null' % (editor, npfile))
        npfile = make_auxilary(base, type_, fmt, '-test')
        # os.system('%s "%s" > /dev/null' % (editor, npfile))
    else:
        sfile = make_style(base, type_, fmt)
        # cmd = '%s "%s" > /dev/null' % (editor, sfile)
        # os.system(cmd)



"""
Provides the |Converter| object which defines the basic mechanism for
converting the objects defined in :mod:`lexor.core.elements`. This
involves using objects derived from the abstract class
|NodeConverter|.

There are a couple of functions that have been written to be used
inside python embeddings. These are:

- :func:`get_converter_namespace`
- :func:`get_lexor_namespace`
- :func:`get_current_node` equivalent to ``__NODE__``
- :func:`echo`
- :func:`include`
- :func:`import_module`

When writing python embeddings there are several special variables
that can be used, one of them as mentioned above is ``__NODE__``
which has the returned value of the function
:func:`get_current_node`. The other two are ``__FILE__`` which takes
in the value of the path of the document which is executing the
python embedding and ``__DIR__`` which is equivalent to
``os.path.dirname(__FILE__)``.

.. |Converter| replace:: :class:`~.Converter`
.. |Parser| replace:: :class:`~lexor.core.parser.Parser`
.. |NodeConverter| replace:: :class:`~.Converter`
.. |Document| replace:: :class:`~lexor.core.elements.Document`
.. |DocFrag| replace:: :class:`~lexor.core.elements.DocumentFragment`
.. |Text| replace:: :class:`~lexor.core.elements.Text`
.. |PI| replace:: :class:`~lexor.core.elements.ProcessingInstruction`

"""
import sys
import os.path as pth
import traceback
from imp import load_source
from cStringIO import StringIO
from lexor.command import config
from lexor.command.lang import get_style_module, map_explanations
LC = sys.modules['lexor.core']


def get_converter_namespace():
    """Many converters may be defined during the conversion of a
    document. In some cases we may need to save references to objects
    in documents. If this is the case, then call this function to
    obtain the namespace where you can save those references. """
    return get_converter_namespace.namespace
if not hasattr(get_converter_namespace, 'namespace'):
    get_converter_namespace.namespace = dict()


class NodeConverter(object):
    """A node converter is an object which determines if the node
    will be copied (default). To avoid copying the node simply
    declare::

        copy = False

    when deriving a node converter. Note that by default, the
    children of the node (if any) will be copied and assigned to the
    parent. To avoid copying the children then set::

        copy_children = False

    """
    copy = True
    copy_children = True

    def __init__(self, converter):
        """A node converter needs to be initialized with a converter
        object. If this method is to be overloaded then make sure
        that it only accepts one parameter: `converter`. This method
        is used by |Converter| and it calls it with itself as the
        parameter. """
        self.converter = converter

    @classmethod
    def start(cls, node):
        """This method gets called only if ``copy`` is set to True
        (default). By overloading this method you have access to the
        converter and the node. You can thus set extra variables in
        the converter or modify the node. **DO NOT** modify any of
        the parents of the node. If there is a need to modify any of
        parents of the node then set a variable in the converter to
        point to the node so that later on in the ``convert``
        function it can be modified. """
        return node

    @classmethod
    def end(cls, node):
        """This method gets called after all the children have been
        copied. Make sure to return the node or the node replacement.
        """
        return node

    def msg(self, code, node, arg=None, uri=None):
        """Send a message to the converter by providing one of the
        error codes defined in the style as well as the position
        where the error took place. The position has the form of a
        list containing the line and column number. Some error codes
        may provide arguments, this can be passed to `arg`. In case
        the error occurred somewhere not in the current document,
        perhaps in a string, then you may provide a new `uri` to
        denote the location."""
        self.converter.msg(self.__module__, code, node, arg, uri)


# pylint: disable=R0903
class BaseLog(object):
    """A simple class to provide messages to a converter. You must
    derive an object from this class in the module which will be
    issuing the messages. For instance::

        class Log(BaseLog):
            pass

    After that you can create a new object and use it in a module::

        log = Log(converter)

    where `converter` is a |Converter| provided to the module. Make
    sure that the module contains the objects ``MSG`` and
    ``MSG_EXPLANATION``."""

    def __init__(self, converter):
        self.converter = converter

    def msg(self, code, arg=None, uri=None):
        """Send a message to the converter. """
        self.converter.msg(self.__module__, code, None, arg, uri)


# The default of 7 attributes for class is too restrictive.
# pylint: disable=R0902
class Converter(object):
    """To see the languages available to the converter see the
    :mod:`lexor.command.lang` module. """

    def __init__(self, fromlang='xml', tolang='xml',
                 style='default', defaults=None):
        """Create a new converter by specifying the language and the
        style in which node objects will be written. """
        if defaults is None:
            defaults = dict()
        self._fromlang = fromlang
        self._tolang = tolang
        self._style = style
        self._nc = None
        self._node_converter = None
        self._convert_func = None
        self._reload = True
        self.style_module = None
        self.doc = list()
        self.log = list()
        self.defaults = defaults

    def __getitem__(self, name):
        """Return the specified |NodeConverter|. """
        return self._node_converter[name]

    @property
    def convert_from(self):
        """The language from which the converter will convert. """
        return self._fromlang

    @convert_from.setter
    def convert_from(self, value):
        """Setter function for convert_from. """
        self._fromlang = value
        self._reload = True

    @property
    def convert_to(self):
        """The language to which the converter will convert. """
        return self._tolang

    @convert_to.setter
    def convert_to(self, value):
        """Setter function for convert_to. """
        self._tolang = value
        self._reload = True

    @property
    def converting_style(self):
        """The converter style. """
        return self._style

    @converting_style.setter
    def converting_style(self, value):
        """Setter function for converting_style. """
        self._style = value
        self._reload = True

    def set(self, fromlang, tolang, style, defaults=None):
        """Sets the languages and styles in one call. """
        if defaults is not None:
            self.defaults = defaults
        self._style = style
        self._tolang = tolang
        self._fromlang = fromlang
        self._reload = True

    def match_info(self, fromlang, tolang, style, defaults=None):
        """Check to see if the converter main information matches."""
        match = True
        if defaults is not None:
            match = False
        elif fromlang not in [self._fromlang]:
            match = False
        elif tolang not in [self._tolang]:
            match = False
        elif style not in [self._style]:
            match = False
        return match

    @property
    def lexor_log(self):
        """The `lexorlog` document. See this document after each
        call to :meth:`convert` to see warnings and errors. """
        return self.log[-1]

    @property
    def document(self):
        """The parsed document. This is a |Document| or |DocFrag|
        created by the :meth:`convert` method. """
        return self.doc[-1]

    def pop(self):
        """Remove the last document and last log document and return
        them. """
        return self.doc.pop(), self.log.pop()

    def convert(self, doc, namespace=False):
        """Convert the |Document| or |DocFrag| doc. """
        if not isinstance(doc, (LC.Document, LC.DocumentFragment)):
            raise TypeError("The node is not a Document or DocumentFragment")
        if self._reload:
            self._set_node_converters(
                self._fromlang, self._tolang, self._style, self.defaults
            )
            self._reload = False
        self.log.append(LC.Document("lexor", "log"))
        self.log[-1].modules = dict()
        self.log[-1].explanation = dict()
        self._convert(doc)
        if hasattr(self.style_module, 'convert'):
            self.style_module.convert(self, self.doc[-1])
        map_explanations(self.log[-1].modules, self.log[-1].explanation)
        if not namespace:
            del self.doc[-1].namespace
        self.doc[-1].lang = self._tolang
        self.doc[-1].style = 'default'
        return self.doc[-1], self.log[-1]

    @staticmethod
    def remove_node(node):
        """Removes the node from the current document it is in.
        Returns the previous sibling is possible, otherwise it
        returns an empty |Text| node. """
        parent = node.parent
        index = node.index
        del node.parent[node.index]
        try:
            if index - 1 > -1:
                return parent[index-1]
            else:
                raise IndexError
        except IndexError:
            parent.append_child('')
        return parent[0]

    # pylint: disable=R0913
    def msg(self, mod_name, code, node, arg=None, uri=None):
        """Provide the name of module issuing the message, the `code`
        number, the node with the error, optional arguments and
        `uri`. This information gets stored in the converters log. """
        if uri is None:
            uri = self.doc[-1].uri_
        if arg is None:
            arg = ()
        wnode = LC.Void('msg')
        wnode['module'] = mod_name
        wnode['code'] = code
        wnode['node_id'] = id(node)
        wnode.node = node
        try:
            wnode['uri'] = node['uri']
            del node['uri']
        except (KeyError, TypeError):
            wnode['uri'] = uri
        wnode['arg'] = arg
        if mod_name not in self.log[-1].modules:
            self.log[-1].modules[mod_name] = sys.modules[mod_name]
        self.log[-1].append_child(wnode)

    def _set_node_converter(self, val):
        """Helper function to create a node converter and store it in
        a dictionary. """
        if isinstance(val, str):
            return self._node_converter[val]
        name = val.__name__
        self._node_converter[name] = val(self)
        return self._node_converter[name]

    def _set_node_converters(self, fromlang, tolang, style, defaults=None):
        """Imports the correct module based on the languages and
        style. """
        self.style_module = get_style_module(
            'converter', fromlang, style, tolang
        )
        name = '%s-converter-%s-%s' % (fromlang, tolang, style)
        config.set_style_cfg(self, name, defaults)
        self._nc = dict()
        self._node_converter = dict()
        if hasattr(self.style_module, 'REPOSITORY'):
            for val in self.style_module.REPOSITORY:
                self._set_node_converter(val)
        self._nc['__default__'] = self._set_node_converter(NodeConverter)
        str_key = list()
        for key, val in self.style_module.MAPPING.iteritems():
            if isinstance(val, str) and val not in self._node_converter:
                str_key.append((key, val))
            else:
                self._nc[key] = self._set_node_converter(val)
        for key, val in str_key:
            self._nc[key] = self._nc[val]

    def _start(self, node):
        """Evaluate the start function of the node converter based
        on the name of the node. """
        return self._nc.get(node.name, self._nc['__default__']).start(node)

    def _end(self, node):
        """Evaluate the end function of the node converter based
        on the name of the node. """
        return self._nc.get(node.name, self._nc['__default__']).end(node)

    def _copy(self, node):
        """Return the copy attribute of the node converter. """
        return self._nc.get(node.name, self._nc['__default__']).copy

    def _copy_children(self, node):
        """Return the copy_children attribute of the node converter. """
        tmp = self._nc.get(node.name, self._nc['__default__']).copy_children
        return tmp and node.child

    def _get_direction(self, crt):
        """Returns the direction in which the traversal should go. """
        if crt.child and self._copy_children(crt):
            return 'd'
        return 'r'

    def _clone_node(self, crt):
        """Clones the node if the node converter assigned to the node
        has the copy property set to True. """
        if self._copy(crt):
            return crt.clone_node()
        return LC.Text('')

    def _convert(self, doc):
        """Main convert function. """
        direction = None
        # A doc needs to be copied by default. You may prohibit
        # to copy the children, but there must be a document.
        crt = doc
        self.doc.append(doc.clone_node())
        self.doc[-1].namespace = dict()
        if hasattr(self.style_module, 'init_conversion'):
            self.style_module.init_conversion(self, self.doc[-1])
        crtcopy = self.doc[-1]
        crtcopy = self._start(crtcopy)
        if self._copy_children(crt):
            direction = 'd'
            root = doc
        else:
            return
        while True:
            if direction is 'd':
                crt = crt.child[0]
                clone = self._clone_node(crt)
                crtcopy.append_child(clone)
            elif direction is 'r':
                if crt.next is None:
                    direction = 'u'
                    continue
                crt = crt.next
                clone = self._clone_node(crt)
                crtcopy.parent.append_child(clone)
            elif direction is 'u':
                crtcopy = self._end(crtcopy.parent)
                crtcopy.normalize()
                if crt.parent is root:
                    break
                if crt.parent.next is None:
                    crt = crt.parent
                    continue
                crt = crt.parent.next
                clone = self._clone_node(crt)
                crtcopy.parent.append_child(clone)
            crtcopy = clone
            if self._copy(crt):
                crtcopy = self._start(crtcopy)
                direction = self._get_direction(crt)
            else:
                direction = 'r'

    def update_log(self, log, after=True):
        """Append the messages from a `log` document to the
        converters log. This removes the children from `log`. """
        modules = log.modules
        explanation = log.explanation
        for mname in modules:
            if mname not in self.log[-1].modules:
                self.log[-1].modules[mname] = modules[mname]
            if mname not in self.log[-1].explanation:
                self.log[-1].explanation[mname] = explanation[mname]
        if after:
            self.log[-1].extend_children(log)
        else:
            self.log[-1].extend_before(0, log)

    # pylint: disable=W0122,E1103
    def exec_python(self, node, id_num, parser, error=True):
        """Executes the contents of the |PI| node. You must provide
        an id number identifying the processing instruction and a
        |Parser| that will parse the output provided by the
        execution. If `error` is True then any errors generated
        during the execution will be appended to the output of the
        document."""
        get_current_node.current.append(node)
        include.converter.append(self)
        namespace = get_lexor_namespace()
        if '__NAMESPACE__' not in namespace:
            namespace['__NAMESPACE__'] = namespace
            namespace['import_module'] = import_module
            namespace['include'] = include
            namespace['echo'] = echo
        namespace['__FILE__'] = pth.realpath(include.converter[-1].doc[-1].uri)
        namespace['__DIR__'] = pth.dirname(namespace['__FILE__'])
        namespace['__NODE__'] = get_current_node()
        original_stdout = sys.stdout
        sys.stdout = StringIO()
        try:
            exec(node.data, namespace)
        except BaseException:
            self.msg(self.__module__, 'E100', node, [id_num])
            if error:
                err_node = LC.Element('python_pi_error')
                err_node['section'] = str(id_num)
                err_node.append_child(
                    LC.CData(traceback.format_exc())
                )
                node.parent.insert_before(node.index, err_node)
        text = sys.stdout.getvalue()
        sys.stdout.close()
        sys.stdout = original_stdout
        parser.parse(text)
        node.parent.extend_before(node.index, parser.doc)
        newnode = Converter.remove_node(node)
        if parser.log:
            self.msg(self.__module__, 'W101', node, [id_num])
            self.update_log(parser.log)
            self.msg(self.__module__, 'W102', node, [id_num])
        get_current_node.current.pop()
        include.converter.pop()
        if include.converter:
            doc = include.converter[-1].doc[-1]
            namespace['__FILE__'] = pth.realpath(doc.uri)
            namespace['__DIR__'] = pth.dirname(namespace['__FILE__'])
            namespace['__NODE__'] = get_current_node()
        else:
            namespace['__FILE__'] = None
            namespace['__DIR__'] = None
            namespace['__NODE__'] = None
        return newnode


def get_lexor_namespace():
    """The execution of python instructions take place in the
    namespace provided by this function."""
    return get_lexor_namespace.namespace
if not hasattr(get_lexor_namespace, 'namespace'):
    get_lexor_namespace.namespace = dict()


def get_current_node():
    """Return the document node containing the python embeddings
    currently being executed. """
    return get_current_node.current[-1]
if not hasattr(get_current_node, 'current'):
    get_current_node.current = list()


def echo(node):
    """Allows the insertion of Nodes generated within python
    embeddings::

        <?python
        comment = PI('!--', 'This is a comment')
        echo(comment)
        ?>

    """
    crt = get_current_node()
    if isinstance(node, str):
        crt.parent.insert_before(crt.index, LC.Text(node))
    elif isinstance(node, LC.Node):
        if node.name == '#document':
            crt.parent.extend_before(crt.index, node)
            return
        crt.parent.insert_before(crt.index, node)
    elif isinstance(node, list):
        for item in node:
            echo(item)
    else:
        while node:
            echo(node[0])


def include(input_file, **keywords):
    """Inserts a file into the current node. Absolute paths may be
    provided as well as relative. When using relative paths the files
    are found relative to the path of the calling document. You may
    use the following keywords:
    
    - parser_style: ``'default'``
    - parser_lang: ``None``
    - parser_defaults: ``None``,
    - convert_style: ``'default'``,
    - convert_from: ``None``,
    - convert_to: ``None``,
    - convert_defaults: ``None``,
    - adopt: ``'true'``
    
    If the keyword ``adopt`` is set to false then a |Document| node
    will be inserted."""
    parent_converter = include.converter[-1]
    if input_file[0] != '/':
        input_file = pth.join(pth.dirname(parent_converter.doc.uri),
                              input_file)
    info = {
        'parser_style': 'default',
        'parser_lang': None,
        'parser_defaults': None,
        'convert_style': 'default',
        'convert_from': None,
        'convert_to': None,
        'convert_defaults': None,
        'adopt': True,
    }
    for key in keywords:
        info[key] = keywords[key]
    if info['parser_lang'] is None:
        path = pth.realpath(input_file)
        name = pth.basename(path)
        name = pth.splitext(name)
        info['parser_lang'] = name[1][1:]
    with open(input_file, 'r') as tmpf:
        text = tmpf.read()
    parser = LC.Parser(info['parser_lang'],
                       info['parser_style'],
                       info['parser_defaults'])
    parser.parse(text, input_file)
    if parser.log:
        parent_converter.update_log(parser.log)
    crt = get_current_node()
    if info['convert_to'] is not None:
        if info['convert_from'] is None:
            info['convert_from'] = info['parser_lang']
        converter = Converter(info['convert_from'],
                              info['convert_to'],
                              info['convert_style'],
                              info['convert_defaults'])
        converter.convert(parser.doc)
        if converter.log:
            parent_converter.update_log(converter.log)
        doc = converter.document
    else:
        doc = parser.doc
    if info['adopt']:
        crt.parent.extend_before(crt.index, doc)
    else:
        crt.parent.insert_before(crt.index, doc)
if not hasattr(include, 'converter'):
    include.converter = list()


def import_module(mod_path, mod_name=None):
    """Return a module from a path. If no name is provided then the
    name of the file loaded will be assigned to the name. When using
    relative paths, it will find the module relative to the file
    executing the python embedding. """
    doc = include.converter[-1].doc[-1]
    if not mod_path.endswith('.py'):
        mod_path += '.py'
    if mod_path[0] != '/':
        mod_path = pth.join(pth.dirname(doc.uri), mod_path)
    if mod_name is None:
        mod_name = pth.basename(mod_path)
    if mod_name.endswith('.py'):
        mod_name = mod_name[:-3]
    return load_source(mod_name, mod_path)


MSG = {
    'E100': 'errors in python processing instruction section `{0}`',
    'W101': '--> begin ?python section `{0}` messages',
    'W102': '--> end ?python section `{0}` messages',
}
MSG_EXPLANATION = [
    """
    - This message is being shown because of E100.

    - The python processing instructions has mistakes. See the
      traceback generated to fix the errors.

    - If the traceback is not shown in the document it may be
      due to the option `error` being off.

""",
    """
    - Python embeddings may generate output to be adapted to the
      document. Such output also needs to be processed. When the
      output generates errors these errors get appended to the
      converter log document.

    - All messages between W101 and W102 are are simply errors of the
      parsed output.

""",
]



class Position(object):
    """Wrapper object around an array of two integers to hold the
    carets position in the document. This can be useful when
    refering to a position in the document in our errors. For
    instance:

        MSG = {
            'E101': 'no more backticks after {0} to match',
        }

        NodeParser.msg('E101', parser.pos, [Position(pos)])

    In this way when an error is display it will display the position
    in whichever format the `Position` object has stored.
    """

    def __init__(self, pos, fmt='{0}:{1:2}'):
        self.pos = list(pos)
        self.fmt = fmt

    def __str__(self):
        return self.fmt.format(*self.pos)

    def __repr__(self):
        return repr(self.pos)

    def set_format(self, fmt='{0}:{1:2}'):
        """Set the display format. To reset to the default value
        call with no arguments.
        """
        self.fmt = fmt

    def shift(self, delta):
        """Shift the position by the amount specified by delta."""
        self.pos[0] += delta[0]
        self.pos[1] += delta[1]



"""to

Execute lexor by transforming a file "to" another language.

"""

import re
import os
import sys
import textwrap
import argparse
from lexor.util.logging import L
from lexor.command import config, LexorError
from lexor.core.parser import Parser
from lexor.core.writer import Writer
from lexor.core.converter import Converter


DEFAULTS = {
    'parse_lang': 'lexor:_',
    'log': 'lexor:log',
    'lang': 'html[_:_]'
}

DESC = """
Transform the inputfile to another language. To see the available
languages use the `lang` command.

examples:

    lexor file.html to markdown[cstyle:wstyle]
    lexor file.html to markdown[cstyle:otherlang.wstyle]
    lexor file.html to html~min,plain,_~
    lexor file.html to html~plain,_~ mk[cstyle:wstyle,cstyle1,cstyle2]

    lexor file.xml to xml~_@tab="****"~

    lexor Makefile to xml --from makefile
    lexor Makefile to xml --from makefile:otherstyle

  Store output to `output.html` and store warnings in `log.html`:
      lexor doc.md > output.html 2> log.html

  Pipe the output from another program:
       cat doc.md | lexor --from markdown

  Write to files without displaying output:
      lexor --quiet --nodisplay --write doc.md

"""


def split_at(delimiter, text, opens='[<(', closes=']>)', quotes='"\''):
    """Custom function to split at commas. Taken from stackoverflow
    http://stackoverflow.com/a/20599372/788553"""
    result = []
    buff = ""
    level = 0
    is_quoted = False
    for char in text:
        if char in delimiter and level == 0 and not is_quoted:
            result.append(buff)
            buff = ""
        else:
            buff += char
            if char in opens:
                level += 1
            elif char in closes:
                level -= 1
            elif char in quotes:
                is_quoted = not is_quoted
    if not buff == "":
        result.append(buff)
    return result


def language_style(lang_str):
    """Parses a language string. In particular, the options --from
    and --log. """
    tmp = split_at(':', lang_str.lower())
    if len(tmp) == 2:
        input_lang = tmp[0]
        input_style = style_parameters(tmp[1])
    else:
        index = tmp[0].find('@')
        if index == -1:
            input_lang = tmp[0]
            input_style = style_parameters('_')
        else:
            input_lang = tmp[0][:index]
            input_style = style_parameters(tmp[0][index:])
    if input_lang == '':
        input_lang = '_'
    if input_style['name'] == '':
        input_style['name'] = '_'
    return input_lang, input_style


def parse_styles(lang_str):
    """Parses a language string. In particular, the options --from
    and --log. """
    tmp = split_at(':', lang_str)
    if len(tmp) == 2:
        cstyle = style_parameters(tmp[0])
        wstyle = style_parameters(tmp[1])
    else:
        cstyle = style_parameters(tmp[0])
        wstyle = style_parameters('_')
    if cstyle['name'] == '':
        cstyle['name'] = '_'
    if wstyle['name'] == '':
        wstyle['name'] = '_'
    return cstyle, wstyle


def style_parameters(style):
    """Parsers a style name along with its parameters. """
    style = split_at('@', style)
    style_dict = {'name': style[0], 'params': dict()}
    for ele in style[1:]:
        try:
            var, val = ele.split('=')
        except ValueError:
            msg = 'style arguments must be of the form @var=val'
            raise argparse.ArgumentTypeError(msg)
        try:
            if val[0] == '[':
                val = val[1:-1]
        except IndexError:
            msg = 'To give empty values try @var=[]'
            raise argparse.ArgumentTypeError(msg)
        style_dict['params'][var] = val
    return style_dict


def input_language(tolang):
    """Parses the tolang argument. """
    type_ = 'w'
    match = re.match("(.+)~(.+)~", tolang)
    if not match:
        type_ = 'c'
        index = tolang.find('[')
        if index == -1:
            if tolang[-1] == ']':
                msg = 'did you mean to write lang[...]?'
                raise argparse.ArgumentTypeError(msg)
            lang_name = tolang
            styles = '_'
        else:
            if tolang[-1] != ']':
                msg = 'must be of the form lang[...] or lang~...~'
                raise argparse.ArgumentTypeError(msg)
            lang_name = tolang[:index]
            styles = tolang[index+1:-1]
    else:
        lang_name, styles = match.groups()
    if type_ == 'w':
        msg = '`:` is not supported in "%s~%s~"' % (lang_name, styles)
        styles = split_at(',', styles)
        for style in styles:
            if ':' in style:
                pass
                # raise argparse.ArgumentTypeError(msg)
        styles = [style_parameters(ele) for ele in styles]
    else:
        styles = [parse_styles(ele) for ele in split_at(',', styles)]
    return type_, lang_name, styles


def add_parser(subp, fclass):
    """Add a parser to the main subparser. """
    tmpp = subp.add_parser('to',
                           help='transform inputfile to another language',
                           formatter_class=fclass,
                           description=textwrap.dedent(DESC))
    tmpp.add_argument('tolang', metavar='lang', nargs='*',
                      type=input_language,
                      help='language to which it will be converted')
    tmpp.add_argument('--from', type=language_style, metavar='FROM',
                      dest='parse_lang',
                      help='language to be parsed in')
    tmpp.add_argument('--log', type=language_style,
                      help='language in which the logs will be written')
    tmpp.add_argument('--write', '-w', action='store_true',
                      help='write to file')
    tmpp.add_argument('--quiet', '-q', action='store_true',
                      help='suppress warning messages')
    tmpp.add_argument('--nodisplay', '-n', action='store_true',
                      help="suppress output")


def get_input(input_file, default='_'):
    """Returns the text to be parsed along with the name assigned to
    that text. The last output is the extension of the file. """
    if input_file is '_':
        L.info('reading input from stdin')
        return sys.stdin.read(), 'STDIN', 'STDIN', default
    L.info('checking existance of `%s`', input_file)
    if not os.path.exists(input_file):
        raise LexorError('`%s` does not exist' % input_file)
    text = open(input_file).read()
    textname = input_file
    path = os.path.realpath(input_file)
    name = os.path.basename(path)
    name = os.path.splitext(name)
    file_name = name[0]
    file_ext = name[1][1:].lower()
    if file_ext == '':
        file_ext = default  # The default language to parse
    return text, textname, file_name, file_ext


def run():
    """Run the command. """
    arg = config.CONFIG['arg']
    cfg = config.get_cfg(['to'])
    text, t_name, f_name, f_ext = get_input(arg.inputfile)

    parse_lang = cfg['to']['parse_lang']
    if isinstance(parse_lang, str):
        parse_lang = language_style(parse_lang)
    if arg.parse_lang is None and f_ext != '_':
        parse_lang = (f_ext, parse_lang[1])
    in_lang, in_style = parse_lang

    log = cfg['to']['log']
    if isinstance(log, str):
        log = language_style(log)
    else:
        default_log = DEFAULTS['log'].split(':')
        if log[0] == '_':
            log = (default_log[0], log[1])
        if log[1]['name'] == '_':
            log[1]['name'] = default_log[1]

    parser = Parser(in_lang, in_style['name'], in_style['params'])
    log_writer = Writer(log[0], log[1]['name'], log[1]['params'])
    try:
        parser.parse(text, t_name)
    except ImportError:
        msg = 'Parsing style not found: [%s:%s]'
        raise LexorError(msg % (in_lang, in_style['name']))
    try:
        write_log(log_writer, parser.log, arg.quiet)
    except ImportError:
        msg = "ERROR: Writing log style not found: [%s:%s]\n"
        raise LexorError(msg % (parser.log.lang, parser.log.style))
    if not arg.tolang:
        arg.tolang.append(input_language(cfg['to']['lang']))
    convert_and_write(f_name, parser, in_lang, log, arg)


def convert_and_write(f_name, parser, in_lang, log, arg):
    """Auxiliary function to reduce the number of branches in run. """
    log_writer = Writer(log[0], log[1]['name'], log[1]['params'])
    writer = Writer()
    converter = Converter()
    param = {
        'parser': parser,
        'converter': converter,
        'writer': writer,
        'log_writer': log_writer,
        'f_name': f_name,
        'in_lang': in_lang,
        'arg': arg
    }
    for (action, lang, styles) in arg.tolang:
        param['styles'] = styles
        param['lang'] = lang
        if action == 'c':
            run_converter(param)
        if action == 'w':
            run_writer(param)


def write_log(writer, log, quiet):
    """Write the log file to stderr. """
    if quiet is False and len(log) > 0:
        writer.write(log, sys.stderr)


def write_document(writer, doc, fname, arg):
    """Auxiliary function for convert_and_write. """
    if arg.nodisplay is False:
        writer.write(doc, sys.stdout)
    if arg.write is True:
        writer.write(doc, fname)


def run_converter(param):
    """Auxiliary function for convert and write. """
    lang = param['lang']
    f_name = param['f_name']
    in_lang = param['in_lang']
    arg = param['arg']
    parser = param['parser']
    writer = param['writer']
    log_writer = param['log_writer']
    for style in param['styles']:
        cstyle = style[0]['name']
        param['converter'].set(in_lang, lang, cstyle, style[0]['params'])
        try:
            doc, log = param['converter'].convert(parser.doc)
        except ImportError:
            msg = 'converting style not found: [%s ==> %s:%s]'
            L.warn(msg % (in_lang, lang, cstyle))
            continue
        wstyle = style[1]['name']
        if '.' in wstyle:
            (lang, wstyle) = wstyle.split('.')
        if wstyle == '_':
            wstyle = 'default'
        writer.set(lang, wstyle, style[1]['params'])
        fname = '%s.%s.%s' % (f_name, wstyle, lang)
        write_log(log_writer, log, arg.quiet)
        try:
            write_document(writer, doc, fname, arg)
        except ImportError:
            msg = 'writing style not found: [%s:%s]'
            L.warn(msg % (lang, wstyle))
            continue


def run_writer(param):
    """Auxiliary function for convert and write. """
    lang = param['lang']
    f_name = param['f_name']
    arg = param['arg']
    parser = param['parser']
    writer = param['writer']
    for style in param['styles']:
        sname = 'default' if style['name'] == '_' else style['name']
        writer.set(lang, style['name'], style['params'])
        fname = '%s.%s.%s' % (f_name, sname, lang)
        try:
            write_document(writer, parser.doc, fname, arg)
        except ImportError:
            msg = 'writing style not found: [%s:%s]'
            L.warn(msg % (lang, style['name']))



"""
Collection of functions to create lexor's command line utility.

"""
import sys
from dateutil import parser
from datetime import datetime
from subprocess import Popen, PIPE


def disp(msg):
    """Print a message to the standard output. """
    sys.stdout.write(msg)


def wdisp(msg):
    """Print a message to the standard error. """
    sys.stderr.write(msg)


def import_mod(name):
    """Return a module by string. """
    mod = __import__(name)
    for sub in name.split(".")[1:]:
        mod = getattr(mod, sub)
    return mod


def exec_cmd(cmd, verbose=False):
    """Run a subprocess and return its output, errors and return code
    when `verbose` is set to False. Otherwise execute the command
    `cmd`. """
    if verbose:
        out = sys.stdout
        err = sys.stderr
    else:
        out = PIPE
        err = PIPE
    process = Popen(cmd, shell=True,
                    universal_newlines=True, executable="/bin/bash",
                    stdout=out, stderr=err)
    out, err = process.communicate()
    return out, err, process.returncode


def date(short=False):
    """Return the current date as a string. """
    if isinstance(short, str):
        now = parser.parse(short)
        return now.strftime("%a %b %d, %Y %r")
    now = datetime.now()
    if not short:
        return now.strftime("%a %b %d, %Y %r")
    return now.strftime("%Y-%m-%d-%H-%M-%S")


class LexorError(Exception):
    """Every known error should be raised via this exception. """
    pass



"""
This module is in charge of providing all the necessary settings to
the rest of the modules in lexor.

The config module comes with global variable ``CONFIG`` which is
meant to be read only. This variable is a dictionary with the
following keys:

- ``path``: The path to the configuration file
- ``name``: The name of the configuration file (lexor.config)
- ``cfg_path``
- ``cfg_user``
- ``arg``

The last three keys are meant to be used for the command line, but
they can come in handy in scripts if we need to specify the location
of the configuration file.

"""
import os
import sys
import argparse
import textwrap
import configparser
import os.path as pth
from lexor.util.logging import L
from lexor.command import import_mod, LexorError, disp


DESC = """
view and edit a configuration file for lexor.

Some actions performed by lexor can be overwritten by using
configuration files.

To see the values that the configuration file can overwrite use the
`defaults` command. This will print a list of the keys and values
lexor uses for the given command.

"""

CONFIG = {
    'path': None,  # read only
    'name': None,  # read only
    'cfg_path': None,  # COMMAND LINE USE ONLY
    'cfg_user': None,  # COMMAND LINE USE ONLY
    'arg': None,  # COMMAND LINE USE ONLY
    'cache': None,  # read_config USE ONLY
}


# pylint: disable=too-few-public-methods
class _ConfigDispAction(argparse.Action):
    """Derived argparse Action class to use when displaying the
    configuration file and location."""
    def __call__(self, parser, namespace, values, option_string=None):
        CONFIG['cfg_user'] = namespace.cfg_user
        CONFIG['cfg_path'] = namespace.cfg_path
        cfg_file = read_config()
        fname = '%s/%s' % (CONFIG['path'], CONFIG['name'])
        disp('lexor configuration file: %s\n' % fname)
        cfg_file.write(sys.stdout)
        exit(0)


def add_parser(subp, fclass):
    """
    .. admonition:: Command Line Utility Function
        :class: warning

        Add a parser to the main sub-parser.
    """
    tmpp = subp.add_parser('config', help='configure lexor',
                           formatter_class=fclass,
                           description=textwrap.dedent(DESC))
    tmpp.add_argument('var', type=str,
                      help='Must be in the form of sec.key')
    tmpp.add_argument('value', type=str, nargs='?', default=None,
                      help='var value')
    tmpp.add_argument('-v', action='store_true',
                      help='print config file location')
    tmpp.add_argument('--display', action=_ConfigDispAction,
                      nargs=0,
                      help='print config file and exit')


def run():
    """
    .. admonition:: Command Line Utility Function
        :class: warning

        Run the command.
    """
    arg = CONFIG['arg']
    cfg_file = read_config()
    try:
        command, var = arg.var.split('.', 1)
    except ValueError:
        raise LexorError("'%s' is not of the form sec.key" % arg.var)
    if arg.v:
        fname = '%s/%s' % (CONFIG['path'], CONFIG['name'])
        disp('lexor configuration file: %s\n' % fname)
    if arg.value is None:
        L.info('displaying value for `%s` and exiting' % var)
        try:
            disp('%s\n' % cfg_file[command][var])
        except KeyError:
            pass
        return
    try:
        cfg_file[command][var] = arg.value
    except KeyError:
        L.info('adding section `%s`' % command)
        cfg_file.add_section(command)
        cfg_file[command][var] = arg.value
    L.info('added key `%s`' % var)
    L.info('writing configuration file')
    write_config(cfg_file)


def read_config(cache=True):
    """Read a configuration file. There are a few ways of specifying
    which configuration file to use.

    - Setting ``CONFIG['cfg_user']`` to ``True`` will read the file
      ``~/.lexor.config``. This variable is set to true in the
      command line by using ``--cfg-user``.

    - Specifying a path via ``CONFIG['cfg_path']`` reads the file
      ``lexor.config`` in the specified path. We can set this
      variable in the command line via the option ``--cfg``.

    - If no path is specified by ``CONFIG['cfg_path']`` then it
      searches for ``lexor.config`` in the current directory. If this
      fails then it will attempt to look for ``lexor.config`` in the
      path specified by the environmental variable
      ``LEXOR_CONFIG_PATH``.

    - If everything else fails then it searches for ``.lexor.config``
      in the home directory.

    This function may raise a :class:`~lexor.command.LexorError`
    exception if the configuration is not found when
    ``CONFIG['cfg_path']`` is set.

    .. note::

        If you must read a specific configuration file via a script
        you can import ``lexor.command.config`` and then set
        ``'cfg_path'``. For instance::

            from lexor.command import config
            config.CONFIG['cfg_path'] = 'path/to/config/file'
            print config.read_config()

        If only the ``CONFIG`` variable is imported then the changes
        will not propagate to the module. That is, if we were to
        do::

            from lexor.command.config import CONFIG
            CONFIG['cfg_path'] = 'path/to/config/file'
            print config.read_config()

        then the module will not be aware of the file since
        ``CONFIG`` is a copy of the ``CONFIG`` variable in the
        ``config`` module.

    """
    if cache and CONFIG['cache'] is not None:
        L.info('reading cached configuration')
        return CONFIG['cache']
    cfg_file = configparser.ConfigParser(allow_no_value=True)
    name = 'lexor.config'
    if CONFIG['cfg_user']:
        path = os.environ['HOME']
        name = '.lexor.config'
    elif CONFIG['cfg_path'] is None:
        path = '.'
        if not pth.exists(name):
            if 'LEXOR_CONFIG_PATH' in os.environ:
                path = os.environ['LEXOR_CONFIG_PATH']
            else:
                path = os.environ['HOME']
                name = '.lexor.config'
    else:
        path = CONFIG['cfg_path']
        if not pth.exists('%s/%s' % (path, name)):
            raise LexorError('%s/%s does not exist.' % (path, name))
    cfg_file.read('%s/%s' % (path, name))
    CONFIG['name'] = name
    CONFIG['path'] = path
    CONFIG['cache'] = cfg_file
    L.info('loaded configuration `%s` from `%s`', name, path)
    return cfg_file


def write_config(cfg_file):
    """Write the configuration file. The input to this function
    should be the object returned by :func:`read_config`. This will
    write the changes made to the configuration in the location
    specified by the ``CONFIG`` global variable in the config module.
    """
    fname = '%s/%s' % (CONFIG['path'], CONFIG['name'])
    with open(fname, 'w') as tmp:
        cfg_file.write(tmp)
    L.info('wrote configuration file `%s`', fname)


def update_single(cfg, name, defaults=None):
    """Update the specified section in configuration (``name``) with
    the ``defaults`` provided. If no defaults are provided then it
    will attempt to look for a lexor command and obtain its defaults
    so that ``cfg`` may be updated."""
    if defaults:
        for var, val in defaults.iteritems():
            cfg[name][var] = pth.expandvars(str(val))
    else:
        try:
            mod = import_mod('lexor.command.%s' % name)
            if hasattr(mod, "DEFAULTS"):
                for var, val in mod.DEFAULTS.iteritems():
                    cfg[name][var] = pth.expandvars(val)
        except ImportError:
            pass


def _update_from_file(cfg, name, cfg_file):
    "Helper function for get_cfg."
    if name in cfg_file:
        for var, val in cfg_file[name].iteritems():
            cfg[name][var] = pth.expandvars(val)


def _update_from_arg(cfg, argdict, key):
    "Helper function for get_cfg."
    for var in cfg[key]:
        if var in argdict and argdict[var] is not None:
            cfg[key][var] = argdict[var]


def get_cfg(names, defaults=None):
    """Obtain settings from the configuration file. Sometimes we may
    wish to only obtain certain sections from the configuration. When
    this is the case we can use this function and we specify a list
    of ``names`` or a single string to obtain specific sections from
    the configuration. We may optionally provide defaults in case we
    wish to override the defaults provided for each of the
    sections."""
    cfg = {
        'lexor': {
            'path': ''
        }
    }
    L.info('getting configuration for: %s' % names)
    cfg_file = read_config()
    if 'lexor' in cfg_file:
        for var, val in cfg_file['lexor'].iteritems():
            cfg['lexor'][var] = pth.expandvars(val)
    cfg['lexor']['root'] = CONFIG['path']
    if isinstance(names, list):
        for name in names:
            cfg[name] = dict()
            update_single(cfg, name)
            _update_from_file(cfg, name, cfg_file)
    else:
        if names != 'lexor':
            cfg[names] = dict()
            update_single(cfg, names, defaults)
            _update_from_file(cfg, names, cfg_file)
    if CONFIG['arg']:
        argdict = vars(CONFIG['arg'])
        if argdict['parser_name'] in cfg:
            _update_from_arg(cfg, argdict, argdict['parser_name'])
        _update_from_arg(cfg, argdict, 'lexor')
    return cfg


def set_style_cfg(obj, name, defaults):
    """Given a |Parser|, |Converter| or |Writer| ``obj``, it sets its
    ``defaults`` attribute to the specified defaults in the
    configuration file or by the user by overwriting values in the
    parameter ``defaults``.

    .. |Parser| replace:: :class:`~lexor.core.parser.Parser`
    .. |Converter| replace:: :class:`~lexor.core.converter.Converter`
    .. |Writer| replace:: :class:`~lexor.core.writer.Writer`

    """
    L.info('configuring %r', obj.__class__)
    obj.defaults = dict()
    if hasattr(obj.style_module, 'DEFAULTS'):
        L.info('setting module defaults for %r', name)
        mod_defaults = obj.style_module.DEFAULTS
        for var, val in mod_defaults.iteritems():
            obj.defaults[var] = pth.expandvars(str(val))
    cfg_file = read_config()
    if name in cfg_file:
        L.info('setting defaults from configuration for %r', name)
        for var, val in cfg_file[name].iteritems():
            obj.defaults[var] = pth.expandvars(val)
    if defaults:
        L.info('overwriting default values from user for %r', name)
        for var, val in defaults.iteritems():
            obj.defaults[var] = val



"""
To use lexor as a module you should explore in detail the packages
provided with lexor. These packages contain many other functions and
information which can help you convert your document in the way you
desire.

:|core|:
    The core of lexor defines basic objects such as ``Document`` and
    provides the main objects that define the functions provided in
    this module.

:|command|:
    This module is in charge of providing all the available commands
    to lexor.

.. |core| replace:: :mod:`~lexor.core`
.. |command| replace:: :mod:`~lexor.command`

----------------------------------------------------------------------

In this module we can find useful functions to quickly parse, convert
and write files without first creating any of the main lexor objects.

"""
import re
import os
import os.path as pth
from sys import stdout
from os.path import realpath, basename, splitext
from lexor.__version__ import get_version
from lexor.command.lang import load_aux
from lexor import core

__all__ = [
    'lexor',
    'parse',
    'read',
    'convert',
    'write',
    'init',
]


def _read_text(src, search=False):
    """Attempt to read a file and return its contents. """
    try:
        return open(src, 'r').read()
    except IOError:
        if search is False:
            return None
    try:
        lexorinputs = os.environ['LEXORINPUTS']
    except KeyError:
        return None
    for directory in lexorinputs.split(':'):
        path = '%s/%s' % (directory, src)
        try:
            return open(path, 'r').read()
        except IOError:
            pass
    return None


def lexor(src, search=False, **keywords):
    """Utility function to parse and convert a file or string
    specified by `src`. If `search` is ``True`` then it will
    attemp to search for `src` in the paths specified by the
    enviromental variable ``$LEXORINPUTS``. The following are all the
    valid keywords and its defaults that this function accepts:

    - parser_style: ``'_'``
    - parser_lang: ``None``
    - parser_defaults: ``None``,
    - convert_style: ``'_'``,
    - convert_from: ``None``,
    - convert_to: ``'html'``,
    - convert_defaults: ``None``,
    - convert: ``'true'``

    Returns the converted ``Document`` object and the log
    ``Document`` containing possible warning or error messages.
    """
    info = {
        'parser_style': '_',
        'parser_lang': None,
        'parser_defaults': None,
        'convert_style': '_',
        'convert_from': None,
        'convert_to': 'html',
        'convert_defaults': None,
        'convert': 'true'
    }
    for key in keywords:
        info[key] = keywords[key]
    text = _read_text(src, search)
    if text is None:
        match = re.match(r'^(\.|/)(.*/)?(?:$|(.+?)(?:(\.[^.]*$)|$))', src)
        if match:
            raise IOError('file `%s` not found' % src)
        else:
            text = src
            if info['parser_lang'] is None:
                info['parser_lang'] = 'lexor'
            src = 'string@0x%x' % id(text)
    elif info['parser_lang'] is None:
        path = pth.realpath(src)
        name = pth.basename(path)
        name = pth.splitext(name)
        info['parser_lang'] = name[1][1:]
    parser = core.Parser(info['parser_lang'],
                         info['parser_style'],
                         info['parser_defaults'])
    parser.parse(text, src)
    if info['convert'] == 'true' and info['convert_to'] is not None:
        if info['convert_from'] is None:
            info['convert_from'] = info['parser_lang']
        converter = core.Converter(info['convert_from'],
                                   info['convert_to'],
                                   info['convert_style'],
                                   info['convert_defaults'])

        converter.convert(parser.doc)
        if parser.log:
            converter.update_log(parser.log, False)
        cdoc = converter.doc.pop()
        clog = converter.log.pop()
    else:
        cdoc = parser.doc
        clog = parser.log
    return cdoc, clog


def parse(text, lang='xml', style='default'):
    """Parse the `text` in the language specified by `lang` and
    return its ``Document`` form and a ``Document`` log containing
    the errors encountered during parsing. """
    parser = core.Parser(lang, style)
    parser.parse(text)
    return parser.document, parser.log


def read(filename, style="default", lang=None):
    """Read and parse a file. If `lang` is not specified then the
    language is assummed from the `filename` extension. Returns the
    ``Document`` form and a ``Document`` log containing the errors
    encountered during parsing. """
    if lang is None:
        path = realpath(filename)
        name = basename(path)
        name = splitext(name)
        lang = name[1][1:]
    with open(filename, 'r') as tmpf:
        text = tmpf.read()
    parser = core.Parser(lang, style)
    parser.parse(text, filename)
    return parser.document, parser.log


def convert(doc, lang=None, style="default"):
    """Convert the ``Document`` `doc` to another language in a given
    style. If the `lang` is not specified then the `doc` is
    tranformed to the same language as the ``Document`` using the
    default style. """
    if lang is None:
        lang = doc.owner.lang
    converter = core.Converter(doc.owner.lang, lang, style)
    converter.convert(doc)
    return converter.document, converter.log


def write(doc, filename=None, mode='w', **options):
    """Write `doc` to a file. To write to the standard output use the
    default parameters, otherwise provide `filename`. If `filename`
    is provided you have the option of especifying the mode: ``'w'``
    or ``'a'``.

    You may also provide a file you may have opened yourself in place
    of filename so that the writer writes to that file, in this case
    the `mode` is ignored.

    The valid `options` depends on the language the document
    specifies. See the ``DEFAULT`` values a particular writer style
    has to obtain the valid options. """
    if doc.owner is None:
        style = 'default'
        lang = 'xml'
    else:
        style = doc.owner.style
        lang = doc.owner.lang
    writer = core.Writer(lang, style)
    if doc.owner is not None and doc.owner.defaults is not None:
        for var, val in doc.owner.defaults.iteritems():
            writer.defaults[var] = os.path.expandvars(str(val))
    for var, val in options.iteritems():
        writer.defaults[var] = os.path.expandvars(str(val))
    if filename:
        writer.write(doc, filename, mode)
    else:
        writer.write(doc, stdout)


def init(**keywords):
    """Every lexor style needs to call the ``init`` function. These
    are the valid keywords to initialize a style:

    - version: ``(major, minor, micro, alpha/beta/rc/final, #)``
    - lang
    - [to_lang]
    - type
    - description
    - [github]: ``{'user', 'repo'}``
    - author: ``{'name', 'email'}``
    - [docs]
    - license
    - path: Must be set to ``__file__``.

    """
    required_keys = [
        'version', 'lang', 'type', 'description', 'author', 'license',
        'path'
    ]
    optional_keys = [
        'to_lang', 'github', 'docs'
    ]
    for key in required_keys:
        if key not in keywords:
            raise NameError('lexor.init requies `%s`' % key)
    info = dict()
    for key in optional_keys:
        info[key] = None
    for key in keywords:
        info[key] = keywords[key]
    info['style'] = splitext(basename(info['path']))[0]
    info['style'] = info['style'].split('-')[0]
    info['ver'] = get_version(info['version'])
    return info



"""Command line use of lexor

To run lexor from the command line do the following:

    python -m lexor ...

Use the option --help for more information.

"""
import sys
import argparse
import textwrap
import os.path as pt
from glob import iglob
from lexor.__version__ import VERSION
from lexor.command import config, import_mod, LexorError
from lexor.util.logging import L


# pylint: disable=W0212
def get_argparse_options(argp):
    """Helper function to preparse the arguments. """
    opt = dict()
    for action in argp._optionals._actions:
        for key in action.option_strings:
            if action.type is None:
                opt[key] = 1
            else:
                opt[key] = 2
    return opt


def preparse_args(argv, argp, subp):
    """Pre-parse the arguments to be able to have a default subparser
    based on the filename provided. """
    opt = get_argparse_options(argp)
    parsers = subp.choices.keys()
    index = 1
    arg = None
    default = 'to'
    try:
        while argv[index] in opt:
            index += opt[argv[index]]
        if index == 1 and argv[index][0] == '-':
            argv.insert(index, 'to')
            argv.insert(index, '_')
            return
        arg = argv[index]
        if arg == 'defaults':
            argv.insert(index, '_')
        if argv[index+1] in parsers:
            return
        if arg not in parsers:
            argv.insert(index+1, default)
    except IndexError:
        if arg not in parsers:
            argv.append(default)
            if arg is None:
                arg = default
    if arg in parsers:
        argv.insert(index, '_')


def parse_options(mod):
    """Interpret the command line inputs and options. """
    desc = """
lexor can perform various commands. Use the help option with a
command for more information.

"""
    ver = "lexor %s" % VERSION
    epi = """
shortcut:

    lexor file.ext lang <==> lexor fle.ext to lang

more info:
  http://jmlopez-rod.github.com/lexor

version:
  lexor %s

""" % VERSION
    raw = argparse.RawDescriptionHelpFormatter
    argp = argparse.ArgumentParser(formatter_class=raw, version=ver,
                                   description=textwrap.dedent(desc),
                                   epilog=textwrap.dedent(epi))
    argp.add_argument('inputfile', type=str, default='_', nargs='?',
                      help='input file to process')
    argp.add_argument('--debug', action='store_true', dest='debug',
                      help='log events')
    argp.add_argument('--lexor-debug', type=str, dest='debug_path',
                      metavar='PATH', default=None,
                      help='diretory to write lexor debug logs')
    argp.add_argument('--cfg', type=str, dest='cfg_path',
                      metavar='CFG_PATH',
                      help='configuration file directory')
    argp.add_argument('--cfg-user', action='store_true',
                      dest='cfg_user',
                      help='select user config file. Overides --cfg')
    subp = argp.add_subparsers(title='subcommands',
                               dest='parser_name',
                               help='additional help',
                               metavar="<command>")

    names = sorted(mod.keys())
    for name in names:
        mod[name].add_parser(subp, raw)
    preparse_args(sys.argv, argp, subp)
    return argp.parse_args()


def run():
    """Run lexor from the command line. """
    mod = dict()
    rootpath = pt.split(pt.abspath(__file__))[0]

    mod_names = [name for name in iglob('%s/command/*.py' % rootpath)]
    for name in mod_names:
        tmp_name = pt.split(name)[1][:-3]
        tmp_mod = import_mod('lexor.command.%s' % tmp_name)
        if hasattr(tmp_mod, 'add_parser'):
            mod[tmp_name] = tmp_mod

    arg = parse_options(mod)

    if arg.debug:
        L.enable()

    config.CONFIG['cfg_path'] = arg.cfg_path
    config.CONFIG['cfg_user'] = arg.cfg_user
    config.CONFIG['arg'] = arg
    try:
        if L.on:
            msg = 'running lexor v%s `%s` command from `%s`'
            L.info(msg, VERSION, arg.parser_name, rootpath)
        mod[arg.parser_name].run()
    except LexorError as err:
        L.error(err.message, exception=err)
    except Exception as err:
        L.error('Unhandled error: ' + err.message, exception=err)

    if arg.debug:
        fp = sys.stderr
        if arg.debug_path:
            try:
                fp = open(pt.join(arg.debug_path, 'lexor.debug'), 'w')
            except IOError as err:
                L.error('invalid debug log directory', exception=err)
        fp.write('[LEXOR DEBUG LOG]\n')
        fp.write('%r\n' % L)
        if arg.debug_path:
            fp.close()


if __name__ == '__main__':
    run()



"""Lexor Test

This module contains various test for lexor.

"""
from __future__ import print_function

import re
import sys
import lexor
import textwrap
from glob import iglob
from nose.tools import eq_
from lexor.command import exec_cmd, wdisp
from lexor.command.lang import LEXOR_PATH, get_style_module
from lexor.core.parser import Parser
from os.path import dirname, exists
from lexor.command import config
from lexor.util.logging import L


RE = re.compile(r'(?P<code>[A-Z][0-9]*|Okay):')
WRAPPER = textwrap.TextWrapper(width=70, break_long_words=False)
DESC = """
This command only takes in one optional parameter. If no parameter
is given then it will attempt to run all the tests available.

If a parameter is given then it may be of one of the following forms:

    lang
    lang.type
    lang.type.style
    lang.converter.tolang.style

The last two forms have the option of being followed by an
specific test name:

    lang.type.style:testname
    lang.converter.tolang.style:testname

"""


def add_parser(subp, fclass):
    """Add a parser to the main subparser. """
    tmpp = subp.add_parser('test', help='test a style',
                           formatter_class=fclass,
                           description=textwrap.dedent(DESC))
    tmpp.add_argument('param', type=str, nargs='?', default='',
                      help='optional parameter to specify the tests')
    tmpp.add_argument('--installed', action='store_true',
                      help='run all installed tests')
    tmpp.add_argument('--verbose', '-v', action='store_true',
                      help='display test messages')


def run():
    """Run command. """
    arg = config.CONFIG['arg']
    cfg = config.get_cfg(['lang', 'develop', 'dependencies'])
    if arg.installed:
        if 'dependencies' in cfg:
            print('Running installed tests ...')
            run_installed(arg.param, cfg, arg.verbose)
        else:
            print('No installed tests to run ...')
    else:
        if 'develop' in cfg:
            print('Running development tests ...')
            run_develop(arg.param, cfg, arg.verbose)
        else:
            print('No development tests to run ...')


def run_develop(param, cfg, verbose):
    """Run develop tests. """
    param = param.split(':')
    testname = '*'
    subtest = ''
    if len(param) > 1:
        testname = param[1]
        if len(param) == 3:
            subtest = ':%s' % param[2]
    keys = [key for key in cfg['develop'] if param[0] in key]
    failed = []
    for key in keys:
        path = cfg['develop'][key][:-3]
        if path[0] != '/':
            path = '%s/%s' % (config.CONFIG['path'], path)
        for pth in iglob('%s/test_%s.py' % (path, testname)):
            cmd = 'nosetests -vs %s%s' % (pth, subtest)
            _, err, _ = exec_cmd(cmd)
            if verbose:
                print(err)
            if 'FAILED' in err:
                failed.append([cmd, err])
    _display_failed(failed)


def run_installed(param, cfg, verbose):
    """Run installed tests. """
    param = param.split(':')
    testname = '*'
    subtest = ''
    if len(param) > 1:
        testname = param[1]
        if len(param) == 3:
            subtest = ':%s' % param[2]
    keys = [key for key in cfg['dependencies'] if key.startswith(param[0])]
    failed = []
    for key in keys:
        name = '/'.join(key.rsplit('.', 1))
        for base in LEXOR_PATH:
            path = '%s/%s' % (base, name)
            L.info('checking %r', path)
            if not exists(path):
                continue
            path = '%s/test_%s.py' % (path, testname)
            for pth in iglob(path):
                cmd = 'nosetests -vs %s%s' % (pth, subtest)
                _, err, _ = exec_cmd(cmd)
                if verbose:
                    print(err)
                if 'FAILED' in err:
                    failed.append([cmd, err])
    _display_failed(failed)


def _display_failed(failed):
    """Helper function. """
    if failed:
        print('FAILED TESTS:')
        print('='*70)
        print('')
    else:
        print('ALL TESTS HAVE PASSED')
    for fail in failed:
        print('COMMAND = %s' % fail[0])
        print('-'*70)
        print(fail[1])


def compare_with(str_obj, expected):
    """Calls ``nose.eq_`` to compare the strings and prints a custom
    message. """
    hline = '_'*60
    msg = "str_obj -->\n%s\n%s\n%s\n\
expected -->\n%s\n%s\n%s\n" % (hline, str_obj, hline,
                               hline, expected, hline)
    eq_(str_obj, expected, msg)


def parse_msg(msg):
    """Obtain the tests embedded inside the messages declared in a
    style. The format of the messages is as follows::

        <tab>[A-Z][0-9]*: <msg>

    or::

        <tab>([A-Z][0-9]*|Okay):
        <tab><tab>msg ...
        <tab><tab>msg continues
        <tab>([A-Z][0-9]*|Okay): msg

    Where ``<tab>`` consists of 4 whitespaces. This function returns
    the message without the tests and a list of tuples of the form
    ``(code, msg)`` along with the message """
    lines = msg.split('\n')
    tests = []
    index = 0
    end = len(lines)
    while index < len(lines):
        match = RE.match(lines[index].strip())
        if match is None:
            index += 1
        else:
            end = index
            break
    while index < len(lines):
        line = lines[index].strip()
        match = RE.match(line)
        if match is not None:
            line = line[match.end():].strip()
            if line == '' and index + 1 < len(lines):
                tmp = lines[index+1]
                if RE.match(tmp[4:]) is None:
                    line += tmp[8:]
                    index += 1
                    while index + 1 < len(lines):
                        tmp = lines[index+1]
                        if RE.match(tmp[4:]) is not None:
                            break
                        line += '\n' + tmp[8:]
                        index += 1
            tests.append((match.group('code'), line))
        index += 1
    return '\n'.join([line[4:] for line in lines[:end]]), tests


def find_failed(tests, lang, style, defaults):
    """Run the tests and return a list of the tests that fail. """
    failed = []
    parser = Parser(lang, style, defaults)
    for test in tests:
        parser.parse(test[1])
        if test[0] == 'Okay':
            if len(parser.log.child) != 0:
                failed.append(test)
        else:
            found = False
            for node in parser.log.child:
                if test[0] == node['code']:
                    found = True
                    break
            if not found:
                failed.append(test)
    return failed


def nose_msg_explanations(lang, type_, style, name, defaults=None):
    """Gather the ``MSG_EXPLANATION`` list and run the tests it
    contains."""
    mod = get_style_module(type_, lang, style)
    mod = sys.modules['%s_%s' % (mod.__name__, name)]
    if not hasattr(mod, 'MSG_EXPLANATION'):
        return
    errors = False
    wdisp('\n')
    for num, msg in enumerate(mod.MSG_EXPLANATION):
        wdisp('MSG_EXPLANATION[%d] ... ' % num)
        msg, tests = parse_msg(msg)
        failed = find_failed(tests, lang, style, defaults)
        err = ['    %s: %r' % (fail[0], fail[1]) for fail in failed]
        if err:
            errors = True
            wdisp('\n%s\n' % '\n'.join(err))
        else:
            wdisp('ok\n')
        err = '\n'.join(err)
    eq_(errors, False, "Errors in MSG_EXPLANATION")
    wdisp('...................... ')


# @deprecated
def print_log(node):
    """Display the error obtained from parsing. """
    pos = node['position']
    sys.stderr.write('\n  Line %d, Column %d ' % (pos[0], pos[1]))
    sys.stderr.write('in %s:\n      ' % node['file'])
    tmp = re.sub(r"\s+", " ", node['message'])
    sys.stderr.write('\n      '.join(WRAPPER.wrap(tmp)))
    sys.stderr.write('\n\n ... ')


# @deprecated
def parse_write(callerfile, in_, out_, style, lang):
    """Provide the filename as the input and the style you wish
    to compare it against. """
    inputfile = "%s/../%s" % (dirname(callerfile), in_)
    with open(inputfile) as tmpf:
        text = tmpf.read()
    sys.stderr.write('\n%s\n%s' % ('='*70, text))
    sys.stderr.write('%s\n' % ('-'*50))
    outputfile = "%s/%s.%s.%s" % (dirname(callerfile), out_, style, lang)
    with open(outputfile) as tmpf:
        text = tmpf.read()
    doc, log = lexor.read(inputfile)
    doc.style = style
    lexor.write(log, sys.stderr)
    sys.stderr.write('\n%s\n' % ('-'*50))
    lexor.write(doc, sys.stderr)
    sys.stderr.write('\n%s\n' % ('='*70))
    return str(doc), text


# @deprecated
def parse_convert_write(callerfile, in_, out_, style, tolang):
    """Provide the filename as the input and the style you wish
    to compare it against. """
    inputfile = "%s/../%s" % (dirname(callerfile), in_)
    with open(inputfile) as tmpf:
        text = tmpf.read()
    sys.stderr.write('\n%s\n%s' % ('='*70, text))
    sys.stderr.write('%s\n' % ('-'*50))
    outputfile = "%s/%s.%s.%s" % (dirname(callerfile), out_, style, tolang)
    with open(outputfile) as tmpf:
        text = tmpf.read()
    doc, log = lexor.read(inputfile)
    doc.style = 'default'
    newdoc, newlog = lexor.convert(doc, tolang, style)
    lexor.write(log, sys.stderr)
    lexor.write(newlog, sys.stderr)
    sys.stderr.write('\n%s\n' % ('-'*50))
    lexor.write(newdoc, sys.stderr)
    sys.stderr.write('\n%s\n' % ('='*70))
    return str(newdoc), text



"""
Package a style along with auxiliary and test files.

"""

import os
import textwrap
import glob
from glob import iglob
from imp import load_source
from zipfile import ZipFile
from lexor.command import config

DEFAULTS = {
    'path': '.'
}

DESC = """
Distribute a style along with auxiliary and test files.

"""


def _style_completer(parsed_args, **_):
    """Return a list of valid files to edit."""
    config.CONFIG['arg'] = parsed_args
    cfg = config.get_cfg('dist', DEFAULTS)
    root = cfg['lexor']['root']
    path = cfg['dist']['path']
    choices = []
    if path[0] in ['/', '.']:
        abspath = path
    else:
        abspath = '%s/%s' % (root, path)
    try:
        if abspath == '.':
            choices.extend(glob.glob('*.py'))
        else:
            choices.extend(glob.glob('%s/*.py' % abspath))
    except OSError:
        pass
    return choices


def add_parser(subp, fclass):
    """
    .. admonition:: Command Line Utility Function
        :class: warning

        Add a parser to the main subparser.
    """
    tmpp = subp.add_parser('dist', help='distribute a style',
                           formatter_class=fclass,
                           description=textwrap.dedent(DESC))
    tmpp.add_argument(
        'style', type=str, help='name of style to distribute'
    ).completer = _style_completer
    tmpp.add_argument('--path', type=str,
                      help='distribution directory')


def run():
    """
    .. admonition:: Command Line Utility Function
        :class: warning

        Run the command.
    """
    arg = config.CONFIG['arg']
    cfg = config.get_cfg('dist', DEFAULTS)
    root = cfg['lexor']['root']
    path = cfg['dist']['path']

    style = arg.style
    if path[0] in ['/', '.']:
        dirpath = path
    else:
        dirpath = '%s/%s' % (root, path)

    if '.py' not in style:
        style = '%s.py' % style
    if not os.path.exists(style):
        error("ERROR: No such file or directory.\n")

    moddir = os.path.splitext(style)[0]
    base, name = os.path.split(moddir)
    if base == '':
        base = '.'

    mod = load_source('tmp_name', style)
    info = mod.INFO
    if info['to_lang']:
        filename = '%s/lexor.%s.%s.%s.%s-%s.zip'
        filename = filename % (dirpath, info['lang'], info['type'],
                               info['to_lang'], info['style'], info['ver'])
    else:
        filename = '%s/lexor.%s.%s.%s-%s.zip'
        filename = filename % (dirpath, info['lang'], info['type'],
                               info['style'], info['ver'])

    warn('Writing %s ...\n' % filename)
    zipf = ZipFile(filename, 'w')
    warn(' including %s\n' % style)
    zipf.write(style)
    for path in iglob('%s/*.py' % moddir):
        warn(' including %s\n' % path)
        zipf.write(path)
    for path in iglob('%s/test_%s/*.py' % (base, name)):
        warn(' including %s\n' % path)
        zipf.write(path)
    zipf.close()



"""Develop

Routine to append a path to the develop section in the configuration
file.

"""

import os
import textwrap
from imp import load_source
from lexor.command import config, LexorError
from lexor.util.logging import L

DESC = """
Append the path to the develop section in a configuration file.

"""


def add_parser(subp, fclass):
    """Add a parser to the main subparser. """
    tmpp = subp.add_parser('develop', help='develop a style',
                           formatter_class=fclass,
                           description=textwrap.dedent(DESC))
    tmpp.add_argument('path', type=str,
                      help='path to the style to develop')


def run():
    """Append the path to the develop section in the configuration
    file. """
    cfg_file = config.read_config()
    arg = config.CONFIG['arg']
    path = os.path.abspath(arg.path)
    if '.py' not in path:
        path = '%s.py' % path
    rel_path = path[len(config.CONFIG['path'])+1:]
    try:
        mod = load_source("tmp-mod", path)
    except IOError:
        msg = 'not a valid module.'
        L.error(msg)
        raise LexorError(msg)
    if not hasattr(mod, 'INFO'):
        msg = 'module does not have `INFO`'
        L.error(msg)
    if mod.INFO['type'] == 'converter':
        key = '%s.%s.%s.%s' % (mod.INFO['lang'], mod.INFO['type'],
                               mod.INFO['to_lang'], mod.INFO['style'])
    else:
        key = '%s.%s.%s' % (mod.INFO['lang'], mod.INFO['type'],
                            mod.INFO['style'])
    if 'develop' in cfg_file:
        if key in cfg_file['develop']:
            if cfg_file['develop'][key] == rel_path:
                msg = '%s is already begin developed from %s'
                msg = msg % (key, path)
                L.error(msg)
                raise LexorError(msg)
        cfg_file['develop'][key] = rel_path
    else:
        cfg_file.add_section('develop')
        cfg_file['develop'][key] = rel_path
    print('%s --> %s' % (key, rel_path))
    config.write_config(cfg_file)
